'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
"""Polars implementation of dataframe-based DFG statistics utilities."""

from typing import Dict, List, Optional, Tuple, Union

import polars as pl

from pm4py.util import constants, xes_constants
from pm4py.util.business_hours import soj_time_business_hours_diff


def _ensure_start_timestamp(
    lf: pl.LazyFrame, start_timestamp_key: str, timestamp_key: str
) -> pl.LazyFrame:
    schema_names = lf.collect_schema().names()
    if start_timestamp_key not in schema_names:
        lf = lf.with_columns(pl.col(timestamp_key).alias(start_timestamp_key))
    return lf


def _ensure_row_index(lf: pl.LazyFrame, column_name: str) -> pl.LazyFrame:
    schema_names = lf.collect_schema().names()
    if column_name in schema_names:
        return lf
    return lf.with_row_count(column_name)


def get_dfg_graph(
    df: pl.LazyFrame,
    measure: str = "frequency",
    activity_key: str = "concept:name",
    case_id_glue: str = "case:concept:name",
    start_timestamp_key: Optional[str] = None,
    timestamp_key: str = "time:timestamp",
    perf_aggregation_key: str = "mean",
    sort_caseid_required: bool = True,
    sort_timestamp_along_case_id: bool = True,
    keep_once_per_case: bool = False,
    window: int = 1,
    business_hours: bool = False,
    business_hours_slot=None,
    workcalendar=constants.DEFAULT_BUSINESS_HOURS_WORKCALENDAR,
    target_activity_key: Optional[str] = None,
    reduce_columns: bool = True,
    cost_attribute: Optional[str] = None,
) -> Union[Dict[Tuple[str, str], int], Dict[Tuple[str, str], float], List[Dict]]:
    """Compute DFG statistics on a Polars LazyFrame."""

    if target_activity_key is None:
        target_activity_key = activity_key

    if start_timestamp_key is None:
        start_timestamp_key = xes_constants.DEFAULT_START_TIMESTAMP_KEY

    schema_names = df.collect_schema().names()
    st_eq_ct = start_timestamp_key == timestamp_key

    if start_timestamp_key not in schema_names:
        df = df.with_columns(pl.col(timestamp_key).alias(start_timestamp_key))
        schema_names = df.collect_schema().names()
        st_eq_ct = True

    needed_columns = {case_id_glue, activity_key, target_activity_key}
    if measure != "frequency" or sort_timestamp_along_case_id:
        needed_columns.update({start_timestamp_key, timestamp_key})
    if measure == "cost" and cost_attribute:
        needed_columns.add(cost_attribute)

    if reduce_columns:
        available = [col for col in schema_names if col in needed_columns]
        df = df.select([pl.col(col) for col in available])
        schema_names = df.collect_schema().names()

    if measure == "cost" and cost_attribute:
        df = df.with_columns(pl.col(cost_attribute).fill_null(0))
        schema_names = df.collect_schema().names()

    if sort_caseid_required:
        if sort_timestamp_along_case_id:
            sort_cols = [case_id_glue]
            if start_timestamp_key in schema_names:
                sort_cols.append(start_timestamp_key)
            if timestamp_key in schema_names:
                sort_cols.append(timestamp_key)
            df = df.sort(sort_cols)
        else:
            df = df.sort(case_id_glue)

    suffix = "_2"
    shift_exprs = [
        pl.col(case_id_glue)
        .shift(-window)
        .over(case_id_glue)
        .alias(case_id_glue + suffix),
        pl.col(target_activity_key)
        .shift(-window)
        .over(case_id_glue)
        .alias(target_activity_key + suffix),
    ]

    if start_timestamp_key in schema_names or start_timestamp_key == timestamp_key:
        shift_exprs.append(
            pl.col(start_timestamp_key)
            .shift(-window)
            .over(case_id_glue)
            .alias(start_timestamp_key + suffix)
        )
    if timestamp_key in schema_names:
        shift_exprs.append(
            pl.col(timestamp_key)
            .shift(-window)
            .over(case_id_glue)
            .alias(timestamp_key + suffix)
        )
    if measure == "cost" and cost_attribute:
        shift_exprs.append(
            pl.col(cost_attribute)
            .shift(-window)
            .over(case_id_glue)
            .alias(cost_attribute + suffix)
        )

    pairs = df.with_columns(shift_exprs)
    pairs = pairs.filter(pl.col(case_id_glue + suffix).is_not_null())
    pairs = pairs.filter(pl.col(target_activity_key + suffix).is_not_null())

    if keep_once_per_case:
        pairs = pairs.unique(
            subset=[case_id_glue, activity_key, target_activity_key + suffix],
            keep="first",
        )

    if measure in {"performance", "both", "cost"}:
        if not st_eq_ct:
            pairs = pairs.with_columns(
                pl.max_horizontal(
                    [pl.col(start_timestamp_key + suffix), pl.col(timestamp_key)]
                ).alias(start_timestamp_key + suffix)
            )

        if business_hours:
            if business_hours_slot is None:
                business_hours_slot = constants.DEFAULT_BUSINESS_HOUR_SLOTS
            pairs = pairs.with_columns(
                pl.struct(
                    [pl.col(timestamp_key), pl.col(start_timestamp_key + suffix)]
                )
                .map_elements(
                    lambda row: soj_time_business_hours_diff(
                        row[timestamp_key],
                        row[start_timestamp_key + suffix],
                        business_hours_slot,
                        workcalendar,
                    ),
                    return_dtype=pl.Float64,
                )
                .alias(constants.DEFAULT_FLOW_TIME)
            )
        else:
            pairs = pairs.with_columns(
                (
                    pl.col(start_timestamp_key + suffix) - pl.col(timestamp_key)
                )
                .dt.total_seconds()
                .alias(constants.DEFAULT_FLOW_TIME)
            )

    group_cols = [activity_key, target_activity_key + suffix]

    freq_dict: Dict[Tuple[str, str], int] = {}
    perf_dict: Dict[Tuple[str, str], Union[float, Dict[str, float], List[float]]] = {}

    if measure in {"frequency", "both"}:
        freq_df = (
            pairs.group_by(group_cols)
            .agg(pl.len().alias("count"))
            .collect()
        )
        freq_dict = {
            (row[group_cols[0]], row[group_cols[1]]): row["count"]
            for row in freq_df.iter_rows(named=True)
            if row[group_cols[0]] is not None and row[group_cols[1]] is not None
        }

    if measure in {"performance", "both", "cost"}:
        if measure == "cost" and cost_attribute:
            value_expr = pl.col(cost_attribute + suffix)
        else:
            value_expr = pl.col(constants.DEFAULT_FLOW_TIME)

        if perf_aggregation_key == "all":
            agg_exprs = [
                value_expr.mean().alias("mean"),
                value_expr.median().alias("median"),
                value_expr.max().alias("max"),
                value_expr.min().alias("min"),
                value_expr.sum().alias("sum"),
                value_expr.std().alias("stdev"),
            ]
            perf_df = pairs.group_by(group_cols).agg(agg_exprs).collect()
            perf_dict = {
                (row[group_cols[0]], row[group_cols[1]]): {
                    "mean": row["mean"],
                    "median": row["median"],
                    "max": row["max"],
                    "min": row["min"],
                    "sum": row["sum"],
                    "stdev": row["stdev"],
                }
                for row in perf_df.iter_rows(named=True)
            }
        elif perf_aggregation_key == "raw_values":
            perf_df = (
                pairs.group_by(group_cols)
                .agg(value_expr.list().alias("raw_values"))
                .collect()
            )
            perf_dict = {
                (row[group_cols[0]], row[group_cols[1]]): row["raw_values"]
                for row in perf_df.iter_rows(named=True)
            }
        else:
            if not hasattr(pl.Expr, perf_aggregation_key):
                raise ValueError(
                    f"Unsupported aggregation key: {perf_aggregation_key}"
                )
            agg_expr = getattr(value_expr, perf_aggregation_key)().alias(
                perf_aggregation_key
            )
            perf_df = (
                pairs.group_by(group_cols).agg(agg_expr).collect()
            )
            perf_dict = {
                (row[group_cols[0]], row[group_cols[1]]): row[perf_aggregation_key]
                for row in perf_df.iter_rows(named=True)
            }

    if measure == "frequency":
        return freq_dict
    if measure in {"performance", "cost"}:
        return perf_dict
    return [freq_dict, perf_dict]


def get_partial_order_dataframe(
    df: pl.LazyFrame,
    start_timestamp_key: Optional[str] = None,
    timestamp_key: str = "time:timestamp",
    case_id_glue: str = "case:concept:name",
    activity_key: str = "concept:name",
    sort_caseid_required: bool = True,
    sort_timestamp_along_case_id: bool = True,
    reduce_dataframe: bool = True,
    keep_first_following: bool = True,
    business_hours: bool = False,
    business_hours_slot=None,
    workcalendar=constants.DEFAULT_BUSINESS_HOURS_WORKCALENDAR,
    event_index: str = constants.DEFAULT_INDEX_KEY,
) -> pl.LazyFrame:
    """Compute partial order dataframe using Polars."""

    if start_timestamp_key is None:
        start_timestamp_key = xes_constants.DEFAULT_START_TIMESTAMP_KEY

    df = _ensure_start_timestamp(df, start_timestamp_key, timestamp_key)

    if reduce_dataframe:
        schema_names = df.collect_schema().names()
        columns = {
            case_id_glue,
            activity_key,
            start_timestamp_key,
            timestamp_key,
        }
        if event_index in schema_names:
            columns.add(event_index)
        df = df.select([pl.col(col) for col in schema_names if col in columns])

    if sort_caseid_required:
        if sort_timestamp_along_case_id:
            df = df.sort([case_id_glue, start_timestamp_key, timestamp_key])
        else:
            df = df.sort(case_id_glue)

    df = _ensure_row_index(df, event_index)

    joined = df.join(df, on=case_id_glue, how="inner", suffix="_2")
    joined = joined.filter(
        pl.col(event_index) < pl.col(event_index + "_2")
    )
    joined = joined.filter(
        pl.col(timestamp_key) <= pl.col(start_timestamp_key + "_2")
    )

    if business_hours:
        if business_hours_slot is None:
            business_hours_slot = constants.DEFAULT_BUSINESS_HOUR_SLOTS
        joined = joined.with_columns(
            pl.struct(
                [pl.col(timestamp_key), pl.col(start_timestamp_key + "_2")]
            )
            .map_elements(
                lambda row: soj_time_business_hours_diff(
                    row[timestamp_key],
                    row[start_timestamp_key + "_2"],
                    business_hours_slot,
                    workcalendar,
                ),
                return_dtype=pl.Float64,
            )
            .alias(constants.DEFAULT_FLOW_TIME)
        )
    else:
        joined = joined.with_columns(
            (
                pl.col(start_timestamp_key + "_2") - pl.col(timestamp_key)
            )
            .dt.total_seconds()
            .alias(constants.DEFAULT_FLOW_TIME)
        )

    if keep_first_following:
        joined = joined.sort([case_id_glue, event_index, event_index + "_2"])
        joined = joined.unique(
            subset=[event_index], keep="first", maintain_order=True
        )

    return joined


def get_concurrent_events_dataframe(
    df: pl.LazyFrame,
    start_timestamp_key: Optional[str] = None,
    timestamp_key: str = "time:timestamp",
    case_id_glue: str = "case:concept:name",
    activity_key: str = "concept:name",
    sort_caseid_required: bool = True,
    sort_timestamp_along_case_id: bool = True,
    reduce_dataframe: bool = True,
    max_start_column: str = "@@max_start_column",
    min_complete_column: str = "@@min_complete_column",
    diff_maxs_minc: str = "@@diff_maxs_minc",
    strict: bool = False,
) -> pl.LazyFrame:
    """Compute concurrency dataframe using Polars."""

    if start_timestamp_key is None:
        start_timestamp_key = xes_constants.DEFAULT_START_TIMESTAMP_KEY

    df = _ensure_start_timestamp(df, start_timestamp_key, timestamp_key)

    if reduce_dataframe:
        columns = [case_id_glue, activity_key, start_timestamp_key, timestamp_key]
        df = df.select([pl.col(col) for col in df.schema if col in columns])

    if sort_caseid_required:
        if sort_timestamp_along_case_id:
            df = df.sort([case_id_glue, start_timestamp_key, timestamp_key])
        else:
            df = df.sort(case_id_glue)

    df = _ensure_row_index(df, constants.DEFAULT_INDEX_KEY)

    joined = df.join(df, on=case_id_glue, how="inner", suffix="_2")
    joined = joined.filter(
        pl.col(constants.DEFAULT_INDEX_KEY)
        < pl.col(constants.DEFAULT_INDEX_KEY + "_2")
    )

    joined = joined.filter(
        pl.col(timestamp_key)
        <= pl.col(start_timestamp_key + "_2")
    )

    joined = joined.with_columns([
        pl.max_horizontal(
            [pl.col(start_timestamp_key), pl.col(start_timestamp_key + "_2")]
        ).alias(max_start_column),
        pl.min_horizontal([pl.col(timestamp_key), pl.col(timestamp_key + "_2")]).alias(
            min_complete_column
        ),
    ])

    joined = joined.with_columns(
        (
            pl.col(min_complete_column) - pl.col(max_start_column)
        )
        .dt.total_seconds()
        .alias(diff_maxs_minc)
    )

    if strict:
        joined = joined.filter(pl.col(diff_maxs_minc) > 0)
    else:
        joined = joined.filter(pl.col(diff_maxs_minc) >= 0)

    return joined.drop(constants.DEFAULT_INDEX_KEY + "_2")



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import Optional, Dict, Any
from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureDFG
from pm4py.algo.discovery.inductive.variants.abc import InductiveMinerFramework
from pm4py.algo.discovery.inductive.variants.instances import IMInstance
from pm4py.algo.discovery.inductive.fall_through.empty_traces import EmptyTracesDFG
from pm4py.objects.process_tree.obj import ProcessTree

class IMD(InductiveMinerFramework[IMDataStructureDFG]):

    def instance(self) -> IMInstance:
        return IMInstance.IMd


    def apply(
        self,
        obj: IMDataStructureDFG,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> ProcessTree:
        # Handle empty traces FIRST (so τ is not lost if a base case/cut triggers)
        empty_traces = EmptyTracesDFG.apply(obj, parameters=parameters)
        if empty_traces is not None:
            return self._recurse(empty_traces[0], empty_traces[1], parameters)
        return super().apply(obj, parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from typing import Optional, Dict, Any
import pandas as pd
from datetime import datetime
from pm4py.util import pandas_utils
import importlib.util


def apply(parameters: Optional[Dict[str, Any]] = None) -> pd.DataFrame:
    """
    Extract a process mining dataframe from all the events recorded in the Windows registry.

    CASE ID (case:concept:name) => name of the computer emitting the events.
    ACTIVITY (concept:name)  => concatenation of the source name of the event and the event identifier
                (see https://learn.microsoft.com/en-us/previous-versions/windows/desktop/eventlogprov/win32-ntlogevent)
    TIMESTAMP (time:timestamp) => timestamp of generation of the event
    RESOURCE (org:resource) => username involved in the event

    Returns
    ----------------
    dataframe
        Pandas dataframe
    """
    if parameters is None:
        parameters = {}

    import win32com.client

    print(
        ":: executing SQL query against the Windows registry. this can take time."
    )

    strComputer = "."
    objWMIService = win32com.client.Dispatch("WbemScripting.SWbemLocator")
    objSWbemServices = objWMIService.ConnectServer(strComputer, "root\\cimv2")
    colItems = objSWbemServices.ExecQuery("Select * from Win32_NTLogEvent")
    events = []

    progress = None
    if importlib.util.find_spec("tqdm"):
        from tqdm.auto import tqdm

        progress = tqdm(
            total=len(colItems), desc="extracting Windows events, progress :: "
        )

    for objItem in colItems:
        events.append(
            {
                "category": str(objItem.Properties_("Category")),
                "categoryString": str(objItem.Properties_("CategoryString")),
                "computerName": str(objItem.Properties_("ComputerName")),
                "eventCode": str(objItem.Properties_("EventCode")),
                "eventIdentifier": str(objItem.Properties_("EventIdentifier")),
                "eventType": str(objItem.Properties_("EventType")),
                "logFile": str(objItem.Properties_("LogFile")),
                "message": str(objItem.Properties_("Message")),
                "recordNumber": str(objItem.Properties_("RecordNumber")),
                "sourceName": str(objItem.Properties_("SourceName")),
                "timeGenerated": datetime.strptime(
                    str(str(objItem.Properties_("TimeGenerated")))
                    .split("+")[0]
                    .split("-")[0],
                    "%Y%m%d%H%M%S.%f",
                ),
                "timeWritten": datetime.strptime(
                    str(str(objItem.Properties_("TimeWritten")))
                    .split("+")[0]
                    .split("-")[0],
                    "%Y%m%d%H%M%S.%f",
                ),
                "type": str(str(objItem.Properties_("Type"))),
                "user": str(str(objItem.Properties_("User"))),
            }
        )
        if progress is not None:
            progress.update()

    if progress is not None:
        progress.close()

    dataframe = pandas_utils.instantiate_dataframe(events)
    dataframe["case:concept:name"] = dataframe["computerName"]
    dataframe["time:timestamp"] = dataframe["timeGenerated"]
    dataframe["concept:name"] = (
        dataframe["sourceName"] + " " + dataframe["eventIdentifier"]
    )
    dataframe["org:resource"] = dataframe["user"]
    dataframe = pandas_utils.insert_index(
        dataframe, "@@index", copy_dataframe=False, reset_index=False
    )
    dataframe = dataframe.sort_values(["time:timestamp", "@@index"])
    dataframe["@@case_index"] = dataframe.groupby(
        "case:concept:name", sort=False
    ).ngroup()
    dataframe = dataframe.sort_values(
        ["@@case_index", "time:timestamp", "@@index"]
    )

    return dataframe



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from typing import Optional, Dict, Any
from enum import Enum
from pm4py.util import exec_utils, pandas_utils
import pandas as pd


class Parameters(Enum):
    CONNECTION_STRING = "connection_string"
    PREFIX = "prefix"


def apply(conn, parameters: Optional[Dict[Any, Any]] = None) -> pd.DataFrame:
    """
    Extracts an event log for the SAP Order-to-Cash process

    Parameters
    ---------------
    conn
        (if provided) connection object to the database (offering cursors)
    parameters
        Parameters of the algorithm, including:
        - Parameters.CONNECTION_STRING => ODBC connection string that is used (if no connection is provided)
        - Parameters.PREFIX => prefix to add to the table names (example SAPSR3.)

    Returns
    ---------------
    dataframe
        Pandas dataframe
    """
    if parameters is None:
        parameters = {}

    import pm4py

    connection_string = exec_utils.get_param_value(
        Parameters.CONNECTION_STRING, parameters, None
    )

    if conn is None:
        import pyodbc

        conn = pyodbc.connect(connection_string)

    prefix = exec_utils.get_param_value(Parameters.PREFIX, parameters, "")

    curs = conn.cursor()

    query = (
        """
SELECT * FROM
(SELECT
    T1.VBELN AS "case:concept:name",
    DECODE(T1.TYPE, 'C', 'Create Sales Document', 'D', 'Create Delivery') AS "concept:name",
    T1.TIMESTAMP AS "time:timestamp",
    T1.USERNAME AS "org:resource"
FROM
    (
        SELECT
            'C' AS TYPE,
            VBAK.VBELN,
            VBAK.ERDAT || ' ' || VBAK.ERZET AS TIMESTAMP,
            VBAK.ERNAM AS USERNAME
        FROM
            """
        + prefix
        + """VBAK
        UNION ALL
        SELECT
            'D' AS TYPE,
            VBFA.VBELV AS VBELN,
            LIKP.ERDAT || ' ' || LIKP.ERZET AS TIMESTAMP,
            LIKP.ERNAM AS USERNAME
        FROM
            """
        + prefix
        + """LIKP
        JOIN """
        + prefix
        + """VBFA ON LIKP.VBELN = VBFA.VBELN
        WHERE
            VBFA.VBTYP_N = 'J'
    ) T1 )
    UNION
    (
    SELECT
    VBFA.VBELN AS "case:concept:name",
    TSTCT.TTEXT AS "concept:name",
    CDHDR.UDATE || ' ' || CDHDR.UTIME AS "time:timestamp",
    CDHDR.USERNAME AS "org:resource"
FROM
    """
        + prefix
        + """VBFA VBFA
    JOIN """
        + prefix
        + """VBAK VBAK ON VBFA.VBELN = VBAK.VBELN
    JOIN """
        + prefix
        + """VBAP VBAP ON VBAK.VBELN = VBAP.VBELN
    JOIN """
        + prefix
        + """CDHDR CDHDR ON VBFA.VBELN = CDHDR.OBJECTID
    JOIN """
        + prefix
        + """TSTCT TSTCT ON CDHDR.TCODE = TSTCT.TCODE
WHERE
    VBFA.VBTYP_N = 'C'
    AND TSTCT.SPRSL = 'E'
    )
    UNION
    (
        SELECT
        C.VBELN AS "case:concept:name",
        'Create Billing' AS "concept:name",
        C.FKDAT AS "time:timestamp",
        C.ERNAM AS "org:resource"
    FROM
        """
        + prefix
        + """VBRK C
    JOIN
        """
        + prefix
        + """VBAK A ON A.VBELN = C.VBELN
    )
    """
    )
    columns = [
        "case:concept:name",
        "concept:name",
        "time:timestamp",
        "org:resource",
    ]

    curs.execute(query)
    dataframe = curs.fetchall()
    dataframe = pandas_utils.instantiate_dataframe_from_records(
        dataframe, columns=columns
    )
    dataframe = pm4py.format_dataframe(dataframe)

    curs.close()
    conn.close()

    return dataframe



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from graphviz import Source
import tempfile
from pm4py.util import exec_utils, constants
from enum import Enum
from typing import Optional, Dict, Any, Union


class Parameters(Enum):
    FORMAT = "format"
    ENABLE_GRAPH_TITLE = "enable_graph_title"
    GRAPH_TITLE = "graph_title"


UNKNOWN_SYMBOL = "&#63;"
XOR_SYMBOL = "&#35;"
PREV_SYMBOL = "&#60;"
SEQUENCE_SYMBOL = "&#62;"
PARALLEL_SYMBOL = "||"


def apply(
    fp1: Dict[str, Any],
    fp2: Dict[str, Any],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Source:
    """
    Visualize a comparison between two footprint tables

    Parameters
    ---------------
    fp1
        Footprints associated to the log (NOT a list)
    fp2
        Footprints associated to the model
    parameters
        Parameters of the algorithm, including:
            - Parameters.FORMAT => Format of the visualization

    Returns
    ---------------
    gviz
        Graphviz object
    """
    if parameters is None:
        parameters = {}

    if type(fp1) is list or type(fp2) is list:
        raise Exception(
            "footprints visualizer does not work on list of footprints!"
        )

    activities1 = sorted(
        list(
            set(x[0] for x in fp1["sequence"])
            .union(set(x[1] for x in fp1["sequence"]))
            .union(set(x[0] for x in fp1["parallel"]))
            .union(set(x[1] for x in fp1["parallel"]))
        )
    )
    activities2 = sorted(
        list(
            set(x[0] for x in fp2["sequence"])
            .union(set(x[1] for x in fp2["sequence"]))
            .union(set(x[0] for x in fp2["parallel"]))
            .union(set(x[1] for x in fp2["parallel"]))
        )
    )
    activities = sorted(list(set(activities1).union(set(activities2))))
    fp_table = {}

    image_format = exec_utils.get_param_value(
        Parameters.FORMAT, parameters, "png"
    )
    enable_graph_title = exec_utils.get_param_value(
        Parameters.ENABLE_GRAPH_TITLE,
        parameters,
        constants.DEFAULT_ENABLE_GRAPH_TITLES,
    )
    graph_title = exec_utils.get_param_value(
        Parameters.GRAPH_TITLE, parameters, "Footprints"
    )

    filename = tempfile.NamedTemporaryFile(suffix=".gv")
    filename.close()

    footprints_table = ["digraph {\n"]

    if enable_graph_title:
        footprints_table.append(
            'label=<<FONT POINT-SIZE="20">'
            + graph_title
            + '</FONT>>;\nlabelloc="top";\n'
        )

    footprints_table.append("tbl [\n")
    footprints_table.append("shape=plaintext\n")
    footprints_table.append("label=<\n")

    footprints_table.append(
        "<table border='0' cellborder='1' color='blue' cellspacing='0'>\n"
    )

    footprints_table.append("<tr><td></td>")
    for act in activities:
        footprints_table.append("<td><b>" + act + "</b></td>")
    footprints_table.append("</tr>\n")

    for a1 in activities:
        footprints_table.append("<tr><td><b>" + a1 + "</b></td>")
        for a2 in activities:
            symb_1 = "?"
            symb_2 = "?"

            if a1 in activities1 and a2 in activities1:
                symb_1 = XOR_SYMBOL
                if (a1, a2) in fp1["parallel"]:
                    symb_1 = PARALLEL_SYMBOL
                elif (a1, a2) in fp1["sequence"]:
                    symb_1 = SEQUENCE_SYMBOL
                elif (a2, a1) in fp1["sequence"]:
                    symb_1 = PREV_SYMBOL

            if a1 in activities2 and a2 in activities2:
                symb_2 = XOR_SYMBOL
                if (a1, a2) in fp2["parallel"]:
                    symb_2 = PARALLEL_SYMBOL
                elif (a1, a2) in fp2["sequence"]:
                    symb_2 = SEQUENCE_SYMBOL
                elif (a2, a1) in fp2["sequence"]:
                    symb_2 = PREV_SYMBOL

            if symb_1 == symb_2:
                footprints_table.append(
                    '<td><font color="black">' + symb_1 + "</font></td>"
                )
            else:
                footprints_table.append(
                    '<td><font color="red">'
                    + symb_1
                    + "&nbsp;&nbsp;"
                    + symb_2
                    + "</font></td>"
                )
        footprints_table.append("</tr>\n")

    footprints_table.append("</table>\n")
    footprints_table.append(">];\n")
    footprints_table.append("}\n")

    footprints_table = "".join(footprints_table)

    gviz = Source(footprints_table, filename=filename.name)
    gviz.format = image_format

    return gviz



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import Optional, Dict, Any
from enum import Enum
from pm4py.util import exec_utils, constants, xes_constants
from pm4py.objects.ocel import constants as ocel_constants
from copy import copy
from pm4py.streaming.stream.live_event_stream import LiveEventStream


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    OCEL_ACTIVITY_KEY = ocel_constants.PARAM_EVENT_ACTIVITY
    OCEL_TIMESTAMP_KEY = ocel_constants.PARAM_EVENT_TIMESTAMP
    OCEL_TYPE_PREFIX = ocel_constants.PARAM_OBJECT_TYPE_PREFIX_EXTENDED


class OcelFlattsDistributor(object):
    def __init__(self, parameters: Optional[Dict[Any, Any]] = None):
        """
        Instantiate the object, "distributing" an OCEL event among all
        the event streams for the "flattened" events.

        Parameters
        -----------------
        parameters
            Parameters of the algorithm, including:
            - Parameters.ACTIVITY_KEY => the activity key to use in the flattening
            - Parameters.TIMESTAMP_KEY => the timestamp key to use in the flattening
            - Parameters.OCEL_ACTIVITY_KEY => the attribute in the OCEL event that is the activity (default: ocel:activity)
            - Parameters.OCEL_TIMESTAMP_KEY => the attribute in the OCEL event that is the timestamp (default: ocel:timestamp)
            - Parameters.OCEL_TYPE_PREFIX => the prefix of the object types in the OCEL (default: ocel:type)
        """
        if parameters is None:
            parameters = {}

        self.activity_key = exec_utils.get_param_value(
            Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
        )
        self.case_id_key = exec_utils.get_param_value(
            Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
        )
        self.timestamp_key = exec_utils.get_param_value(
            Parameters.TIMESTAMP_KEY,
            parameters,
            xes_constants.DEFAULT_TIMESTAMP_KEY,
        )
        self.ocel_activity = exec_utils.get_param_value(
            Parameters.OCEL_ACTIVITY_KEY,
            parameters,
            ocel_constants.DEFAULT_EVENT_ACTIVITY,
        )
        self.ocel_timestamp = exec_utils.get_param_value(
            Parameters.OCEL_TIMESTAMP_KEY,
            parameters,
            ocel_constants.DEFAULT_EVENT_TIMESTAMP,
        )

        self.ot_prefix = exec_utils.get_param_value(
            Parameters.OCEL_TYPE_PREFIX,
            parameters,
            ocel_constants.DEFAULT_OBJECT_TYPE_PREFIX_EXTENDED,
        )
        self.flattened_stream_listeners = {}

    def register(self, object_type: str, live_event_stream: LiveEventStream):
        """
        Register a new event stream (listener) for a given object type.

        Parameters
        -----------------
        object_type
            Given object type
        live_event_stream
            Live event stream
        """
        if object_type not in self.flattened_stream_listeners:
            self.flattened_stream_listeners[object_type] = []
        self.flattened_stream_listeners[object_type].append(live_event_stream)

    def append(self, event: Dict[str, Any]):
        """
        Flattens an OCEL among all the available object types, and send its flattening to each
        corresponding event stream.

        Parameters
        -------------
        event
            OCEL event (obtained for example using the ocel_iterator)
        """
        base_event = {
            x: y for x, y in event.items() if not x.startswith(self.ot_prefix)
        }
        base_event[self.activity_key] = base_event[self.ocel_activity]
        base_event[self.timestamp_key] = base_event[self.ocel_timestamp]
        del base_event[self.ocel_activity]
        del base_event[self.ocel_timestamp]

        ev_objects = {
            x.split(self.ot_prefix)[1]: y
            for x, y in event.items()
            if x.startswith(self.ot_prefix)
        }

        for ot in ev_objects:
            if ot in self.flattened_stream_listeners:
                for obj in ev_objects[ot]:
                    fl_ev = copy(base_event)
                    fl_ev[self.case_id_key] = obj
                    for listener in self.flattened_stream_listeners[ot]:
                        listener.append(fl_ev)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.bpmn.obj import BPMN


def reduce_xor_gateways(bpmn_graph, parameters=None):
    """
    Reduces the number of XOR gateways in the diagram

    Parameters
    ------------
    bpmn_graph
        BPMN graph
    parameters
        Parameters

    Returns
    ------------
    bpmn_graph
        (possibly reduced) BPMN graph
    """
    if parameters is None:
        parameters = {}

    changed = True
    while changed:
        changed = False
        outgoing_edges = None
        incoming_edges = None
        outgoing_edges = {}
        incoming_edges = {}

        for flow in bpmn_graph.get_flows():
            source = flow.get_source()
            target = flow.get_target()

            if source not in outgoing_edges:
                outgoing_edges[source] = set()
            outgoing_edges[source].add(flow)

            if target not in incoming_edges:
                incoming_edges[target] = set()
            incoming_edges[target].add(flow)

        nodes = list(bpmn_graph.get_nodes())
        for node in nodes:
            if isinstance(node, BPMN.ExclusiveGateway):
                if (
                    node in outgoing_edges
                    and node in incoming_edges
                    and len(outgoing_edges[node]) == 1
                    and len(incoming_edges[node]) == 1
                ):
                    changed = True
                    source_node = None
                    target_node = None
                    for flow in incoming_edges[node]:
                        source_node = flow.get_source()
                        if flow in bpmn_graph.get_flows():
                            bpmn_graph.remove_flow(flow)
                    for flow in outgoing_edges[node]:
                        target_node = flow.get_target()
                        if flow in bpmn_graph.get_flows():
                            bpmn_graph.remove_flow(flow)
                    if node in bpmn_graph.get_nodes():
                        bpmn_graph.remove_node(node)
                    bpmn_graph.add_flow(
                        BPMN.SequenceFlow(source_node, target_node)
                    )
                    break

    return bpmn_graph


def apply(bpmn_graph, parameters=None):
    """
    Reduce the complexity of a BPMN graph by removing useless elements

    Parameters
    ------------
    bpmn_graph
        BPMN graph
    parameters
        Parameters

    Returns
    ------------
    bpmn_graph
        (possibly reduced) BPMN graph
    """
    if parameters is None:
        parameters = {}

    bpmn_graph = reduce_xor_gateways(bpmn_graph, parameters=parameters)
    return bpmn_graph



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
__doc__ = """
The ``pm4py.org`` module contains organizational analysis techniques offered in ``pm4py``.
"""

from typing import Union

import pandas as pd

from pm4py.objects.log.obj import EventLog, EventStream
from pm4py.util.pandas_utils import (
    check_is_pandas_dataframe,
    check_pandas_dataframe_columns,
)
from pm4py.utils import get_properties, __event_log_deprecation_warning
from pm4py.objects.org.sna.obj import SNA
from pm4py.objects.org.roles.obj import Role
from pm4py.util import xes_constants
from typing import Dict, Tuple, Any, List


def discover_handover_of_work_network(
    log: Union[EventLog, pd.DataFrame],
    beta=0,
    resource_key: str = "org:resource",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name",
) -> SNA:
    """
    Calculates the handover of work network of the event log.

    The handover of work network is essentially the Directly-Follows Graph (DFG) of the event log, but using the
    resource as the nodes of the graph instead of activities.
    As such, resource information should be present in the event log.

    :param log: Event log or Pandas DataFrame.
    :param beta: Beta parameter for the Handover metric.
    :param resource_key: Attribute to be used for the resource.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as case identifier.

    .. code-block:: python3

        import pm4py

        metric = pm4py.discover_handover_of_work_network(
            dataframe,
            beta=0,
            resource_key='org:resource',
            timestamp_key='time:timestamp',
            case_id_key='case:concept:name'
        )
    """
    __event_log_deprecation_warning(log)

    from pm4py.algo.organizational_mining.sna import algorithm as sna

    parameters = get_properties(
        log,
        resource_key=resource_key,
        timestamp_key=timestamp_key,
        case_id_key=case_id_key,
    )
    parameters["beta"] = beta
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(
            log, timestamp_key=timestamp_key, case_id_key=case_id_key
        )
        return sna.apply(
            log, variant=sna.Variants.HANDOVER_PANDAS, parameters=parameters
        )
    else:
        return sna.apply(
            log, variant=sna.Variants.HANDOVER_LOG, parameters=parameters
        )


def discover_working_together_network(
    log: Union[EventLog, pd.DataFrame],
    resource_key: str = "org:resource",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name",
) -> SNA:
    """
    Calculates the working together network of the process.
    Two resource nodes are connected in the graph if the resources collaborate on an instance of the process.

    :param log: Event log or Pandas DataFrame.
    :param resource_key: Attribute to be used for the resource.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as case identifier.

    .. code-block:: python3

        import pm4py

        metric = pm4py.discover_working_together_network(
            dataframe,
            resource_key='org:resource',
            timestamp_key='time:timestamp',
            case_id_key='case:concept:name'
        )
    """
    __event_log_deprecation_warning(log)

    properties = get_properties(
        log,
        resource_key=resource_key,
        timestamp_key=timestamp_key,
        case_id_key=case_id_key,
    )

    from pm4py.algo.organizational_mining.sna import algorithm as sna

    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(
            log, timestamp_key=timestamp_key, case_id_key=case_id_key
        )
        return sna.apply(
            log,
            variant=sna.Variants.WORKING_TOGETHER_PANDAS,
            parameters=properties,
        )
    else:
        return sna.apply(
            log,
            variant=sna.Variants.WORKING_TOGETHER_LOG,
            parameters=properties,
        )


def discover_activity_based_resource_similarity(
    log: Union[EventLog, pd.DataFrame],
    activity_key: str = "concept:name",
    resource_key: str = "org:resource",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name",
) -> SNA:
    """
    Calculates similarity between the resources in the event log based on their activity profiles.

    :param log: Event log or Pandas DataFrame.
    :param activity_key: Attribute to be used for the activity.
    :param resource_key: Attribute to be used for the resource.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as case identifier.

    .. code-block:: python3

        import pm4py

        act_res_sim = pm4py.discover_activity_based_resource_similarity(
            dataframe,
            resource_key='org:resource',
            activity_key='concept:name',
            timestamp_key='time:timestamp',
            case_id_key='case:concept:name'
        )
    """
    __event_log_deprecation_warning(log)

    properties = get_properties(
        log,
        activity_key=activity_key,
        resource_key=resource_key,
        timestamp_key=timestamp_key,
        case_id_key=case_id_key,
    )

    from pm4py.algo.organizational_mining.sna import algorithm as sna

    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(
            log,
            activity_key=activity_key,
            timestamp_key=timestamp_key,
            case_id_key=case_id_key,
        )
        return sna.apply(
            log,
            variant=sna.Variants.JOINTACTIVITIES_PANDAS,
            parameters=properties,
        )
    else:
        return sna.apply(
            log,
            variant=sna.Variants.JOINTACTIVITIES_LOG,
            parameters=properties,
        )


def discover_subcontracting_network(
    log: Union[EventLog, pd.DataFrame],
    n=2,
    resource_key: str = "org:resource",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name",
) -> SNA:
    """
    Calculates the subcontracting network of the process.

    :param log: Event log or Pandas DataFrame.
    :param n: N parameter for the Subcontracting metric.
    :param resource_key: Attribute to be used for the resource.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as case identifier.

    .. code-block:: python3

        import pm4py

        metric = pm4py.discover_subcontracting_network(
            dataframe,
            n=2,
            resource_key='org:resource',
            timestamp_key='time:timestamp',
            case_id_key='case:concept:name'
        )
    """
    __event_log_deprecation_warning(log)

    from pm4py.algo.organizational_mining.sna import algorithm as sna

    parameters = get_properties(
        log,
        resource_key=resource_key,
        timestamp_key=timestamp_key,
        case_id_key=case_id_key,
    )
    parameters["n"] = n
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(
            log, timestamp_key=timestamp_key, case_id_key=case_id_key
        )
        return sna.apply(
            log,
            variant=sna.Variants.SUBCONTRACTING_PANDAS,
            parameters=parameters,
        )
    else:
        return sna.apply(
            log, variant=sna.Variants.SUBCONTRACTING_LOG, parameters=parameters
        )


def discover_organizational_roles(
    log: Union[EventLog, pd.DataFrame],
    activity_key: str = "concept:name",
    resource_key: str = "org:resource",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name",
) -> List[Role]:
    """
    Mines the organizational roles.

    A role is a set of activities in the log that are executed by a similar (multi)set of resources. Hence, it is a specific function within the organization. Grouping the activities into roles can help:

    Reference paper:
    Burattin, Andrea, Alessandro Sperduti, and Marco Veluscek. “Business models enhancement through discovery of roles.” 2013 IEEE Symposium on Computational Intelligence and Data Mining (CIDM). IEEE, 2013.

    :param log: Event log or Pandas DataFrame.
    :param activity_key: Attribute to be used for the activity.
    :param resource_key: Attribute to be used for the resource.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as case identifier.

    .. code-block:: python3

        import pm4py

        roles = pm4py.discover_organizational_roles(
            dataframe,
            resource_key='org:resource',
            activity_key='concept:name',
            timestamp_key='time:timestamp',
            case_id_key='case:concept:name'
        )
    """
    __event_log_deprecation_warning(log)

    properties = get_properties(
        log,
        activity_key=activity_key,
        resource_key=resource_key,
        timestamp_key=timestamp_key,
        case_id_key=case_id_key,
    )

    from pm4py.algo.organizational_mining.roles import algorithm as roles

    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(
            log,
            activity_key=activity_key,
            timestamp_key=timestamp_key,
            case_id_key=case_id_key,
        )
        return roles.apply(
            log, variant=roles.Variants.PANDAS, parameters=properties
        )
    else:
        return roles.apply(
            log, variant=roles.Variants.LOG, parameters=properties
        )


def discover_network_analysis(
    log: Union[pd.DataFrame, EventLog, EventStream],
    out_column: str,
    in_column: str,
    node_column_source: str,
    node_column_target: str,
    edge_column: str,
    edge_reference: str = "_out",
    performance: bool = False,
    sorting_column: str = xes_constants.DEFAULT_TIMESTAMP_KEY,
    timestamp_column: str = xes_constants.DEFAULT_TIMESTAMP_KEY,
) -> Dict[Tuple[str, str], Dict[str, Any]]:
    """
    Performs a network analysis of the log based on the provided parameters.

    Classical social network analysis methods are based on the order of events within a case. For example, the Handover of Work metric considers the directly-follows relationships between resources during the execution of a case. An edge is added between two resources if such a relationship occurs.

    Real-life scenarios may be more complicated. Firstly, it is difficult to collect events within the same case without encountering convergence/divergence issues (see the first section of the OCEL part). Secondly, the type of relationship may also be important. For example, the relationship between two resources may be more efficient if the activity executed is liked by the resources rather than disliked.

    The network analysis introduced here generalizes some existing social network analysis metrics, making them independent of the case notion and allowing the construction of a multigraph instead of a simple graph.

    We assume events are linked by signals. An event emits a signal (contained in one attribute of the event) that is assumed to be received by other events (also containing this attribute) that follow the first event in the log. We assume there is an OUT attribute (of the event) that is identical to the IN attribute (of the other events).

    When collecting this information, we can build the network analysis graph:
    - The source node of the relationship is determined by aggregating the `node_column_source` attribute.
    - The target node of the relationship is determined by aggregating the `node_column_target` attribute.
    - The type of edge is determined by aggregating the `edge_column` attribute.
    - The network analysis graph can be annotated with frequency or performance information.

    The output is a multigraph.
    Two events EV1 and EV2 in the log are connected (independently of the case notion) based on having
    EV1.OUT_COLUMN = EV2.IN_COLUMN.
    Then, an aggregation is applied on the pair of events (NODE_COLUMN) to obtain the connected nodes.
    The edges between these nodes are aggregated based on some property of the *source* event (`edge_column`).

    :param log: Event log, Pandas DataFrame, or EventStream.
    :param out_column: The source column of the link (default: the case identifier; events of the same case are linked).
    :param in_column: The target column of the link (default: the case identifier; events of the same case are linked).
    :param node_column_source: The attribute to be used for defining the source node (default: the resource of the log, "org:resource").
    :param node_column_target: The attribute to be used for defining the target node (default: the resource of the log, "org:resource").
    :param edge_column: The attribute to be used for defining the edge (default: the activity of the log, "concept:name").
    :param edge_reference: Determines if the edge attribute should be picked from the source event. Values: "_out" => the source event; "_in" => the target event.
    :param performance: Boolean value that enables performance calculation on the edges of the network analysis.
    :param sorting_column: The column to be used for sorting the log before performing the network analysis (default: "time:timestamp").
    :param timestamp_column: The column to be used as timestamp for performance-related analysis (default: "time:timestamp").
    :rtype: ``Dict[Tuple[str, str], Dict[str, Any]]``

    .. code-block:: python3

        import pm4py

        net_ana = pm4py.discover_network_analysis(
            dataframe,
            out_column='case:concept:name',
            in_column='case:concept:name',
            node_column_source='org:resource',
            node_column_target='org:resource',
            edge_column='concept:name',
            edge_reference='_out',
            performance=False,
            sorting_column='time:timestamp',
            timestamp_column='time:timestamp'
        )
    """
    __event_log_deprecation_warning(log)

    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log)

    from pm4py.algo.organizational_mining.network_analysis.variants import (
        dataframe,
    )

    parameters = {}
    parameters[dataframe.Parameters.OUT_COLUMN] = out_column
    parameters[dataframe.Parameters.IN_COLUMN] = in_column
    parameters[dataframe.Parameters.NODE_COLUMN_SOURCE] = node_column_source
    parameters[dataframe.Parameters.NODE_COLUMN_TARGET] = node_column_target
    parameters[dataframe.Parameters.EDGE_COLUMN] = edge_column
    parameters[dataframe.Parameters.EDGE_REFERENCE] = edge_reference
    parameters[dataframe.Parameters.SORTING_COLUMN] = sorting_column
    parameters[dataframe.Parameters.TIMESTAMP_KEY] = timestamp_column
    parameters[dataframe.Parameters.INCLUDE_PERFORMANCE] = performance

    from pm4py.algo.organizational_mining.network_analysis import (
        algorithm as network_analysis,
    )

    return network_analysis.apply(log, parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any

import pandas as pd
from lxml import etree, objectify

from pm4py.objects.ocel import constants
from pm4py.objects.ocel.obj import OCEL
from pm4py.objects.ocel.util import filtering_utils
from pm4py.util import exec_utils, dt_parsing, pandas_utils
from pm4py.objects.ocel.util import ocel_consistency


class Parameters(Enum):
    EVENT_ID = constants.PARAM_EVENT_ID
    EVENT_ACTIVITY = constants.PARAM_EVENT_ACTIVITY
    EVENT_TIMESTAMP = constants.PARAM_EVENT_TIMESTAMP
    OBJECT_ID = constants.PARAM_OBJECT_ID
    OBJECT_TYPE = constants.PARAM_OBJECT_TYPE
    INTERNAL_INDEX = constants.PARAM_INTERNAL_INDEX
    QUALIFIER = constants.PARAM_QUALIFIER
    CHANGED_FIELD = constants.PARAM_CHNGD_FIELD
    ENCODING = "encoding"


def parse_xml(value, tag_str_lower, parser):
    if "float" in tag_str_lower:
        return float(value)
    elif "date" in tag_str_lower:
        return parser.apply(value)
    return str(value)


def apply(file_path: str, parameters: Optional[Dict[Any, Any]] = None) -> OCEL:
    """
    Imports an object-centric event log from a XNK-OCEL file, using LXML

    Parameters
    -----------------
    file_path
        Path to the XML-OCEL file
    parameters
        Parameters of the algorithm, including:
        - Parameters.EVENT_ID
        - Parameters.EVENT_ACTIVITY
        - Parameters.EVENT_TIMESTAMP
        - Parameters.OBJECT_ID
        - Parameters.OBJECT_TYPE
        - Parameters.INTERNAL_INDEX

    Returns
    ------------------
    ocel
        Object-centric event log
    """
    if parameters is None:
        parameters = {}

    events = []
    relations = []
    objects = []
    object_changes = []
    o2o = []
    obj_type_dict = {}

    encoding = exec_utils.get_param_value(
        Parameters.ENCODING, parameters, None
    )

    event_id = exec_utils.get_param_value(
        Parameters.EVENT_ID, parameters, constants.DEFAULT_EVENT_ID
    )
    event_activity = exec_utils.get_param_value(
        Parameters.EVENT_ACTIVITY, parameters, constants.DEFAULT_EVENT_ACTIVITY
    )
    event_timestamp = exec_utils.get_param_value(
        Parameters.EVENT_TIMESTAMP,
        parameters,
        constants.DEFAULT_EVENT_TIMESTAMP,
    )
    object_id = exec_utils.get_param_value(
        Parameters.OBJECT_ID, parameters, constants.DEFAULT_OBJECT_ID
    )
    object_type = exec_utils.get_param_value(
        Parameters.OBJECT_TYPE, parameters, constants.DEFAULT_OBJECT_TYPE
    )
    internal_index = exec_utils.get_param_value(
        Parameters.INTERNAL_INDEX, parameters, constants.DEFAULT_INTERNAL_INDEX
    )
    qualifier_field = exec_utils.get_param_value(
        Parameters.QUALIFIER, parameters, constants.DEFAULT_QUALIFIER
    )
    changed_field = exec_utils.get_param_value(
        Parameters.CHANGED_FIELD, parameters, constants.DEFAULT_CHNGD_FIELD
    )

    date_parser = dt_parsing.parser.get()

    parser = etree.XMLParser(remove_comments=True, encoding=encoding)

    F = open(file_path, "rb")
    tree = objectify.parse(F, parser=parser)
    F.close()

    root = tree.getroot()

    for child in root:
        if child.tag.lower().endswith("events"):
            for event in child:
                eve_id = None
                eve_activity = None
                eve_timestamp = None
                eve_omap = {}
                eve_vmap = {}
                for child2 in event:
                    if child2.get("key") == "id":
                        eve_id = child2.get("value")
                    elif child2.get("key") == "timestamp":
                        eve_timestamp = parse_xml(
                            child2.get("value"),
                            child2.tag.lower(),
                            date_parser,
                        )
                    elif child2.get("key") == "activity":
                        eve_activity = child2.get("value")
                    elif child2.get("key") == "omap":
                        for child3 in child2:
                            objref = child3.get("value")
                            qualifier = (
                                child3.get("qualifier")
                                if "qualifier" in child3.keys()
                                else None
                            )
                            eve_omap[objref] = qualifier
                    elif child2.get("key") == "vmap":
                        for child3 in child2:
                            key = child3.get("key")
                            value = parse_xml(
                                child3.get("value"),
                                child3.tag.lower(),
                                date_parser,
                            )
                            eve_vmap[key] = value

                event_dict = {
                    event_id: eve_id,
                    event_activity: eve_activity,
                    event_timestamp: eve_timestamp,
                }
                for k, v in eve_vmap.items():
                    event_dict[k] = v
                events.append(event_dict)

                for obj in eve_omap:
                    rel_dict = {
                        event_id: eve_id,
                        event_activity: eve_activity,
                        event_timestamp: eve_timestamp,
                        object_id: obj,
                        qualifier_field: eve_omap[obj],
                    }
                    relations.append(rel_dict)
        elif child.tag.lower().endswith("objects"):
            for object in child:
                obj_id = None
                obj_type = None
                obj_ovmap = []
                for child2 in object:
                    if child2.get("key") == "id":
                        obj_id = child2.get("value")
                    elif child2.get("key") == "type":
                        obj_type = child2.get("value")
                    elif child2.get("key") == "ovmap":
                        for child3 in child2:
                            key = child3.get("key")
                            value = parse_xml(
                                child3.get("value"),
                                child3.tag.lower(),
                                date_parser,
                            )
                            timestamp = (
                                child3.get("timestamp")
                                if "timestamp" in child3.keys()
                                else None
                            )
                            obj_ovmap.append((key, value, timestamp))
                dct = {object_id: obj_id, object_type: obj_type}
                for el in obj_ovmap:
                    if el[0] not in dct:
                        dct[el[0]] = el[1]
                    else:
                        this_dct = {object_id: obj_id, object_type: obj_type}
                        this_dct[el[0]] = el[1]
                        this_dct[event_timestamp] = date_parser.apply(el[2])
                        this_dct[changed_field] = el[0]
                        object_changes.append(this_dct)
                objects.append(dct)
                obj_type_dict[obj_id] = obj_type
        elif child.tag.lower().endswith("o2o"):
            for rel in child:
                source = rel.get("source")
                target = rel.get("target")
                qualifier = rel.get("qualifier")
                o2o.append(
                    {
                        object_id: source,
                        object_id + "_2": target,
                        qualifier_field: qualifier,
                    }
                )

    for rel in relations:
        rel[object_type] = obj_type_dict[rel[object_id]]

    events = pandas_utils.instantiate_dataframe(events) if events else None
    objects = pandas_utils.instantiate_dataframe(objects) if objects else None
    relations = (
        pandas_utils.instantiate_dataframe(relations) if relations else None
    )
    o2o = pandas_utils.instantiate_dataframe(o2o) if o2o else None
    object_changes = (
        pandas_utils.instantiate_dataframe(object_changes)
        if object_changes
        else None
    )

    events = pandas_utils.insert_index(
        events, internal_index, reset_index=False, copy_dataframe=False
    )
    relations = pandas_utils.insert_index(
        relations, internal_index, reset_index=False, copy_dataframe=False
    )

    events = events.sort_values([event_timestamp, internal_index])
    relations = relations.sort_values([event_timestamp, internal_index])

    del events[internal_index]
    del relations[internal_index]

    globals = {}

    ocel = OCEL(
        events=events,
        objects=objects,
        relations=relations,
        globals=globals,
        o2o=o2o,
        object_changes=object_changes,
        parameters=parameters,
    )
    ocel = ocel_consistency.apply(ocel, parameters=parameters)
    ocel = filtering_utils.propagate_relations_filtering(ocel)

    return ocel



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.log.obj import EventLog, Trace, EventStream
from pm4py.util import xes_constants as xes
from pm4py.objects.conversion.log import converter as log_converter


def sort_timestamp_trace(
    trace, timestamp_key=xes.DEFAULT_TIMESTAMP_KEY, reverse_sort=False
):
    """
    Sort a trace based on timestamp key

    Parameters
    -----------
    trace
        Trace
    timestamp_key
        Timestamp key
    reverse_sort
        If true, reverses the direction in which the sort is done (ascending)

    Returns
    -----------
    trace
        Sorted trace
    """
    events = sorted(
        trace._list, key=lambda x: x[timestamp_key], reverse=reverse_sort
    )
    new_trace = Trace(events, attributes=trace.attributes)
    return new_trace


def sort_timestamp_stream(
    event_log, timestamp_key=xes.DEFAULT_TIMESTAMP_KEY, reverse_sort=False
):
    """
    Sort an event log based on timestamp key

    Parameters
    -----------
    event_log
        Event log
    timestamp_key
        Timestamp key
    reverse_sort
        If true, reverses the direction in which the sort is done (ascending)

    Returns
    -----------
    event_log
        Sorted event log
    """
    events = sorted(
        event_log._list, key=lambda x: x[timestamp_key], reverse=reverse_sort
    )
    new_stream = EventStream(
        events,
        attributes=event_log.attributes,
        extensions=event_log.extensions,
        omni_present=event_log.omni_present,
        classifiers=event_log.classifiers,
        properties=event_log.properties,
    )
    return new_stream


def sort_timestamp_log(
    event_log, timestamp_key=xes.DEFAULT_TIMESTAMP_KEY, reverse_sort=False
):
    """
    Sort a log based on timestamp key

    Parameters
    -----------
    event_log
        Log
    timestamp_key
        Timestamp key
    reverse_sort
        If true, reverses the direction in which the sort is done (ascending)

    Returns
    -----------
    log
        Sorted log
    """
    event_log = log_converter.apply(
        event_log, variant=log_converter.Variants.TO_EVENT_LOG
    )

    new_log = EventLog(
        attributes=event_log.attributes,
        extensions=event_log.extensions,
        omni_present=event_log.omni_present,
        classifiers=event_log.classifiers,
        properties=event_log.properties,
    )
    for trace in event_log:
        if trace:
            new_log.append(
                sort_timestamp_trace(
                    trace,
                    timestamp_key=timestamp_key,
                    reverse_sort=reverse_sort,
                )
            )
    new_log._list.sort(key=lambda x: x[0][timestamp_key], reverse=reverse_sort)

    return new_log


def sort_timestamp(
    log, timestamp_key=xes.DEFAULT_TIMESTAMP_KEY, reverse_sort=False
):
    """
    Sort a log based on timestamp key

    Parameters
    -----------
    log
        Trace/Event log
    timestamp_key
        Timestamp key
    reverse_sort
        If true, reverses the direction in which the sort is done (ascending)

    Returns
    -----------
    log
        Sorted Trace/Event log
    """
    if type(log) is EventLog:
        return sort_timestamp_log(
            log, timestamp_key=timestamp_key, reverse_sort=reverse_sort
        )
    return sort_timestamp_stream(
        log, timestamp_key=timestamp_key, reverse_sort=reverse_sort
    )


def sort_lambda_log(event_log, sort_function, reverse=False):
    """
    Sort a log based on a lambda expression

    Parameters
    ------------
    event_log
        Log
    sort_function
        Sort function
    reverse
        Boolean (sort by reverse order)

    Returns
    ------------
    new_log
        Sorted log
    """
    event_log = log_converter.apply(
        event_log, variant=log_converter.Variants.TO_EVENT_LOG
    )

    traces = sorted(event_log._list, key=sort_function, reverse=reverse)
    new_log = EventLog(
        traces,
        attributes=event_log.attributes,
        extensions=event_log.extensions,
        omni_present=event_log.omni_present,
        classifiers=event_log.classifiers,
        properties=event_log.properties,
    )

    return new_log


def sort_lambda_stream(event_log, sort_function, reverse=False):
    """
    Sort a stream based on a lambda expression

    Parameters
    ------------
    event_log
        Stream
    sort_function
        Sort function
    reverse
        Boolean (sort by reverse order)

    Returns
    ------------
    stream
        Sorted stream
    """
    events = sorted(event_log._list, key=sort_function, reverse=reverse)
    new_stream = EventStream(
        events,
        attributes=event_log.attributes,
        extensions=event_log.extensions,
        omni_present=event_log.omni_present,
        classifiers=event_log.classifiers,
        properties=event_log.properties,
    )

    return new_stream


def sort_lambda(log, sort_function, reverse=False):
    """
    Sort a log based on lambda expression

    Parameters
    -------------
    log
        Log
    sort_function
        Sort function
    reverse
        Boolean (sort by reverse order)

    Returns
    -------------
    log
        Sorted log
    """
    if type(log) is EventLog:
        return sort_lambda_log(log, sort_function, reverse=reverse)
    return sort_lambda_stream(log, sort_function, reverse=reverse)



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.conformance.alignments.decomposed.variants import (
    recompos_maximal,
)
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any, Union
from pm4py.objects.log.obj import EventLog
from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.util import typing
import pandas as pd


class Variants(Enum):
    RECOMPOS_MAXIMAL = recompos_maximal


VERSIONS = {Variants.RECOMPOS_MAXIMAL}


def apply(
    log: Union[EventLog, pd.DataFrame],
    net: PetriNet,
    im: Marking,
    fm: Marking,
    variant=Variants.RECOMPOS_MAXIMAL,
    parameters: Optional[Dict[Any, Any]] = None,
) -> typing.ListAlignments:
    """
    Apply the recomposition alignment approach
    to a log and a Petri net performing decomposition

    Parameters
    --------------
    log
        Event log
    net
        Petri net
    im
        Initial marking
    fm
        Final marking
    variant
        Variant of the algorithm, possible values:
            - Variants.RECOMPOS_MAXIMAL
    parameters
        Parameters of the algorithm

    Returns
    --------------
    aligned_traces
        For each trace, return its alignment
    """
    return exec_utils.get_variant(variant).apply(
        log, net, im, fm, parameters=parameters
    )



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from pm4py.util import exec_utils
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any, Union, Tuple, Set
from pm4py.algo.conformance.ocel.otg.variants import graph_comparison


class Variants(Enum):
    GRAPH_COMPARISON = graph_comparison


def apply(real: Union[OCEL,
                      Tuple[Set[str],
                            Dict[Tuple[str,
                                       str,
                                       str],
                                 int]]],
          normative: Tuple[Set[str],
                           Dict[Tuple[str,
                                str,
                                str],
                                int]],
          variant=Variants.GRAPH_COMPARISON,
          parameters: Optional[Dict[Any,
                                    Any]] = None) -> Dict[str,
                                                          Any]:
    """
    Applies OTG-based conformance checking between a 'real' object (OCEL or OTG) and a 'normative' OTG.

    Published in: https://publications.rwth-aachen.de/record/1014107

    Parameters
    -----------------
    real
        Real object (OCEL or OTG)
    normative
        Normative OTG
    variant
        Variant of the algorithm to be used (default: Variants.GRAPH_COMPARISON)
    parameters
        Variant-specific parameters

    Returns
    -----------------
    conf_diagn
        Diagnostics dictionary
    """
    return exec_utils.get_variant(variant).apply(real, normative, parameters)



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util.constants import CASE_CONCEPT_NAME
from pm4py.util.xes_constants import DEFAULT_NAME_KEY
from pm4py.util.xes_constants import DEFAULT_TIMESTAMP_KEY
from pm4py.util.constants import PARAMETER_CONSTANT_ATTRIBUTE_KEY
from pm4py.util.constants import PARAMETER_CONSTANT_CASEID_KEY
from pm4py.util.constants import PARAMETER_CONSTANT_TIMESTAMP_KEY
from pm4py.util.constants import DEFAULT_VARIANT_SEP
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any, Union, Tuple, List
import polars as pl
import sys


class Parameters(Enum):
    CASE_ID_KEY = PARAMETER_CONSTANT_CASEID_KEY
    ATTRIBUTE_KEY = PARAMETER_CONSTANT_ATTRIBUTE_KEY
    TIMESTAMP_KEY = PARAMETER_CONSTANT_TIMESTAMP_KEY
    TARGET_ATTRIBUTE_KEY = "target_attribute_key"
    DECREASING_FACTOR = "decreasingFactor"
    POSITIVE = "positive"
    MIN_PERFORMANCE = "min_performance"
    MAX_PERFORMANCE = "max_performance"


def apply(
    df: pl.LazyFrame,
    paths: List[Tuple[str, str]],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Apply a filter on traces containing / not containing a path

    Parameters
    ----------
    df
        LazyFrame
    paths
        Paths to filter on
    parameters
        Possible parameters of the algorithm, including:
            Parameters.CASE_ID_KEY -> Case ID column in the LazyFrame
            Parameters.ATTRIBUTE_KEY -> Attribute we want to filter
            Parameters.POSITIVE -> Specifies if the filter should be applied including traces (positive=True)
            or excluding traces (positive=False)
    Returns
    ----------
    df
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}
    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, DEFAULT_NAME_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, DEFAULT_TIMESTAMP_KEY
    )
    target_attribute_key = exec_utils.get_param_value(
        Parameters.TARGET_ATTRIBUTE_KEY, parameters, attribute_key
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )

    # Convert paths to string format
    paths_str = [path[0] + DEFAULT_VARIANT_SEP + path[1] for path in paths]

    # Sort by case and timestamp
    df = df.sort([case_id_glue, timestamp_key])

    # Get next activity using shift
    df = df.with_columns([
        pl.col(case_id_glue).alias("case_next"),
        pl.col(target_attribute_key).shift(-1).over(case_id_glue).alias("next_activity")
    ])

    # Create path column
    df = df.with_columns(
        pl.when(pl.col("case_next") == pl.col(case_id_glue))
        .then(
            pl.concat_str([
                pl.col(attribute_key),
                pl.lit(DEFAULT_VARIANT_SEP),
                pl.col("next_activity")
            ])
        )
        .otherwise(None)
        .alias("@@path")
    )

    # Get cases with matching paths
    matching_cases = (
        df.filter(pl.col("@@path").is_in(paths_str))
        .select(case_id_glue)
        .unique()
    )

    # Apply positive/negative filter
    if positive:
        ret = df.join(matching_cases, on=case_id_glue, how="inner")
    else:
        ret = df.join(matching_cases, on=case_id_glue, how="anti")

    # Clean up helper columns
    ret = ret.drop(["case_next", "next_activity", "@@path"])

    return ret


def apply_performance(
    df: pl.LazyFrame,
    provided_path: Tuple[str, str],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Filters the cases of a LazyFrame where there is at least one occurrence of the provided path
    occurring in the defined timedelta range.

    Parameters
    ----------
    df
        LazyFrame
    provided_path
        Path to filter on
    parameters
        Possible parameters of the algorithm, including:
            Parameters.CASE_ID_KEY -> Case ID column in the LazyFrame
            Parameters.ATTRIBUTE_KEY -> Attribute we want to filter
            Parameters.TIMESTAMP_KEY -> Attribute identifying the timestamp in the log
            Parameters.POSITIVE -> Specifies if the filter should be applied including traces (positive=True)
            or excluding traces (positive=False)
            Parameters.MIN_PERFORMANCE -> Minimal allowed performance of the provided path
            Parameters.MAX_PERFORMANCE -> Maximal allowed performance of the provided path

    Returns
    ----------
    df
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}
    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, DEFAULT_NAME_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, DEFAULT_TIMESTAMP_KEY
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )
    provided_path_str = provided_path[0] + DEFAULT_VARIANT_SEP + provided_path[1]
    min_performance = exec_utils.get_param_value(
        Parameters.MIN_PERFORMANCE, parameters, 0
    )
    max_performance = exec_utils.get_param_value(
        Parameters.MAX_PERFORMANCE, parameters, sys.maxsize
    )

    # Sort by case and timestamp
    df = df.sort([case_id_glue, timestamp_key])

    # Get next activity and timestamp using shift
    df = df.with_columns([
        pl.col(case_id_glue).alias("case_next"),
        pl.col(attribute_key).shift(-1).over(case_id_glue).alias("next_activity"),
        pl.col(timestamp_key).shift(-1).over(case_id_glue).alias("next_timestamp")
    ])

    # Create path column and calculate duration
    df = df.with_columns([
        pl.when(pl.col("case_next") == pl.col(case_id_glue))
        .then(
            pl.concat_str([
                pl.col(attribute_key),
                pl.lit(DEFAULT_VARIANT_SEP),
                pl.col("next_activity")
            ])
        )
        .otherwise(None)
        .alias("@@path"),
        pl.when(pl.col("case_next") == pl.col(case_id_glue))
        .then(
            (pl.col("next_timestamp") - pl.col(timestamp_key)).dt.total_seconds()
        )
        .otherwise(None)
        .alias("@@duration")
    ])

    # Get cases with matching paths within performance bounds
    matching_cases = (
        df.filter(
            (pl.col("@@path") == provided_path_str) &
            (pl.col("@@duration") >= min_performance) &
            (pl.col("@@duration") <= max_performance)
        )
        .select(case_id_glue)
        .unique()
    )

    # Apply positive/negative filter
    if positive:
        ret = df.join(matching_cases, on=case_id_glue, how="inner")
    else:
        ret = df.join(matching_cases, on=case_id_glue, how="anti")

    # Clean up helper columns
    ret = ret.drop(["case_next", "next_activity", "next_timestamp", "@@path", "@@duration"])

    return ret


'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import matplotlib
from copy import copy

from pm4py.visualization.graphs.util import common
from pm4py.util import exec_utils, constants
from enum import Enum
from typing import Optional, Dict, Any, Union, List


class Parameters(Enum):
    TITLE = "title"
    FORMAT = "format"
    X_AXIS = "x_axis"
    Y_AXIS = "y_axis"
    PYPLOT_PLOT_KWARGS = "pylot_plot_kwargs"
    TRANSPARENT = "transparent"


TIMESTAMP_LABEL = "Date"
DENSITY_LABEL = "Density"
GRAPH_DEFAULT_TITLE = "Events per Time"


def apply_plot(
    x: List[float],
    y: List[float],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> str:
    """
    Plot (non-logarithmic way) the graph with axis values contained in x and y

    Parameters
    ------------
    x
        Values for x-axis
    y
        Values for y-axis
    parameters
        Parameters of the algorithm, including:
            Parameters.FORMAT -> Format of the target image
            Parameters.TITLE -> Title of the image

    Returns
    ------------
    temp_file_name
        Representation temporary file name
    """
    if parameters is None:
        parameters = {}

    format = exec_utils.get_param_value(Parameters.FORMAT, parameters, "png")
    title = exec_utils.get_param_value(
        Parameters.TITLE, parameters, GRAPH_DEFAULT_TITLE
    )
    x_axis = exec_utils.get_param_value(
        Parameters.X_AXIS, parameters, TIMESTAMP_LABEL
    )
    y_axis = exec_utils.get_param_value(
        Parameters.Y_AXIS, parameters, DENSITY_LABEL
    )
    pyplot_plot_kwargs = exec_utils.get_param_value(
        Parameters.PYPLOT_PLOT_KWARGS, parameters, {}
    )
    is_transp = exec_utils.get_param_value(
        Parameters.TRANSPARENT,
        parameters,
        True if constants.DEFAULT_BGCOLOR == "transparent" else False,
    )

    filename = common.get_temp_file_name(format)

    current_backend = copy(matplotlib.get_backend())
    matplotlib.use("Agg")
    from matplotlib import pyplot

    pyplot.clf()
    pyplot.plot(x, y, **pyplot_plot_kwargs)
    pyplot.xlabel(x_axis)
    pyplot.ylabel(y_axis)
    pyplot.title(title)
    pyplot.xticks(rotation=90)
    pyplot.savefig(filename, bbox_inches="tight", transparent=is_transp)
    pyplot.clf()

    matplotlib.use(current_backend)

    return filename


def apply_semilogx(
    x: List[float],
    y: List[float],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> str:
    """
    Plot (semi-logarithmic way) the graph with axis values contained in x and y

    Parameters
    ------------
    x
        Values for x-axis
    y
        Values for y-axis
    parameters
        Parameters of the algorithm, including:
            Parameters.FORMAT -> Format of the target image
            Parameters.TITLE -> Title of the image

    Returns
    ------------
    temp_file_name
        Representation temporary file name
    """
    if parameters is None:
        parameters = {}

    format = exec_utils.get_param_value(Parameters.FORMAT, parameters, "png")
    title = exec_utils.get_param_value(
        Parameters.TITLE, parameters, GRAPH_DEFAULT_TITLE
    )
    pyplot_plot_kwargs = exec_utils.get_param_value(
        Parameters.PYPLOT_PLOT_KWARGS, parameters, {}
    )
    is_transp = exec_utils.get_param_value(
        Parameters.TRANSPARENT,
        parameters,
        True if constants.DEFAULT_BGCOLOR == "transparent" else False,
    )

    filename = common.get_temp_file_name(format)

    current_backend = copy(matplotlib.get_backend())
    matplotlib.use("Agg")
    from matplotlib import pyplot

    pyplot.clf()
    pyplot.semilogx(x, y, **pyplot_plot_kwargs)
    pyplot.xlabel(TIMESTAMP_LABEL)
    pyplot.ylabel(DENSITY_LABEL)
    pyplot.title(title)
    pyplot.xticks(rotation=90)
    pyplot.savefig(filename, bbox_inches="tight", transparent=is_transp)
    pyplot.clf()

    matplotlib.use(current_backend)

    return filename



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum

from pm4py.objects.conversion.log import converter
from pm4py.objects.log.util import sorting
from pm4py.util import exec_utils, constants, xes_constants
from typing import Optional, Dict, Any, Union, Tuple
from pm4py.objects.log.obj import EventLog


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    KEEP_FIRST_FOLLOWING = "keep_first_following"


def apply(
    interval_log: EventLog,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Dict[Tuple[str, str], int]:
    if parameters is None:
        parameters = {}

    interval_log = converter.apply(
        interval_log,
        variant=converter.Variants.TO_EVENT_LOG,
        parameters=parameters,
    )

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    keep_first_following = exec_utils.get_param_value(
        Parameters.KEEP_FIRST_FOLLOWING, parameters, False
    )

    ret_dict = {}
    for trace in interval_log:
        sorted_trace = sorting.sort_timestamp_trace(trace, start_timestamp_key)
        i = 0
        while i < len(sorted_trace):
            act1 = sorted_trace[i][activity_key]
            tc1 = sorted_trace[i][timestamp_key]
            j = i + 1
            while j < len(sorted_trace):
                ts2 = sorted_trace[j][start_timestamp_key]
                act2 = sorted_trace[j][activity_key]
                if tc1 <= ts2:
                    tup = (act1, act2)
                    if tup not in ret_dict:
                        ret_dict[tup] = 0
                    ret_dict[tup] = ret_dict[tup] + 1
                    if keep_first_following:
                        break
                j = j + 1
            i = i + 1

    return ret_dict



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import networkx as nx
from enum import Enum
from typing import Optional, Dict, Any
from pm4py.util import exec_utils, constants
import importlib.util
from copy import copy


class Parameters(Enum):
    SHOW_PROGRESS_BAR = "show_progress_bar"


def get_default_nx_environment():
    return nx


DEFAULT_NX_ENVIRONMENT = get_default_nx_environment()


def Graph(*args, **kwargs):
    return DEFAULT_NX_ENVIRONMENT.Graph(*args, **kwargs)


def DiGraph(*args, **kwargs):
    return DEFAULT_NX_ENVIRONMENT.DiGraph(*args, **kwargs)


def MultiGraph(*args, **kwargs):
    return DEFAULT_NX_ENVIRONMENT.MultiGraph(*args, **kwargs)


def MultiDiGraph(*args, **kwargs):
    return DEFAULT_NX_ENVIRONMENT.MultiDiGraph(*args, **kwargs)


def ancestors(*args, **kwargs):
    return DEFAULT_NX_ENVIRONMENT.ancestors(*args, **kwargs)


def descendants(*args, **kwargs):
    return DEFAULT_NX_ENVIRONMENT.descendants(*args, **kwargs)


def connected_components(*args, **kwargs):
    return DEFAULT_NX_ENVIRONMENT.connected_components(*args, **kwargs)


def bfs_tree(*args, **kwargs):
    return DEFAULT_NX_ENVIRONMENT.bfs_tree(*args, **kwargs)


def contracted_nodes(*args, **kwargs):
    return DEFAULT_NX_ENVIRONMENT.contracted_nodes(*args, **kwargs)


def shortest_path(*args, **kwargs):
    return DEFAULT_NX_ENVIRONMENT.shortest_path(*args, **kwargs)


def strongly_connected_components(*args, **kwargs):
    return DEFAULT_NX_ENVIRONMENT.strongly_connected_components(
        *args, **kwargs
    )


def has_path(*args, **kwargs):
    return DEFAULT_NX_ENVIRONMENT.has_path(*args, **kwargs)


def is_strongly_connected(*args, **kwargs):
    return DEFAULT_NX_ENVIRONMENT.is_strongly_connected(*args, **kwargs)


def all_pairs_shortest_path(*args, **kwargs):
    return DEFAULT_NX_ENVIRONMENT.all_pairs_shortest_path(*args, **kwargs)


def all_pairs_dijkstra(*args, **kwargs):
    return DEFAULT_NX_ENVIRONMENT.all_pairs_dijkstra(*args, **kwargs)


def find_cliques(*args, **kwargs):
    return DEFAULT_NX_ENVIRONMENT.find_cliques(*args, **kwargs)


def degree_centrality(*args, **kwargs):
    return DEFAULT_NX_ENVIRONMENT.degree_centrality(*args, **kwargs)


def greedy_modularity_communities(*args, **kwargs):
    return DEFAULT_NX_ENVIRONMENT.algorithms.community.greedy_modularity_communities(
        *args, **kwargs)


def maximum_flow_value(*args, **kwargs):
    return DEFAULT_NX_ENVIRONMENT.maximum_flow_value(*args, **kwargs)


def minimum_weight_full_matching(*args, **kwargs):
    return DEFAULT_NX_ENVIRONMENT.bipartite.minimum_weight_full_matching(
        *args, **kwargs
    )


def Edmonds(*args, **kwargs):
    return DEFAULT_NX_ENVIRONMENT.algorithms.tree.Edmonds(*args, **kwargs)


def __format_attrs(attributes0: Dict[str, Any]) -> Dict[str, Any]:
    """
    Internal method to format properties.
    """
    attributes = {}

    keys = list(attributes0.keys())

    for k0 in keys:
        v = attributes0[k0]
        t = str(type(v)).lower()

        k = k0
        if "time" in t:
            v = v.isoformat()
            attributes[k] = v
        elif "float" in t or "double" in t:
            attributes[k] = v
        elif "str" in t:
            attributes[k] = v
        else:
            attributes[k] = str(v)

    return attributes


def neo4j_upload(
    nx_graph: nx.DiGraph,
    session,
    clean_db: bool = True,
    parameters: Optional[Dict[Any, Any]] = None,
):
    """
    Uploads a NetworkX DiGraph obtained from a traditional/object-centric event log to a Neo4J session

    Parameters
    ---------------
    nx_graph
        NetworkX graph
    session
        Neo4J session
    clean_db
        Cleans the database before uploading
    parameters
        Other optional parameters of the method, including:
        - Parameters.SHOW_PROGRESS_BAR => shows the percentage of nodes/edges uploaded to Neo4J
    """
    if parameters is None:
        parameters = {}

    show_progress_bar = exec_utils.get_param_value(
        Parameters.SHOW_PROGRESS_BAR, parameters, constants.SHOW_PROGRESS_BAR
    )

    if clean_db:
        session.run("MATCH (n) DETACH DELETE n")

    nodes = list(nx_graph.nodes)
    nodes_progress = None
    edges = list(nx_graph.edges)
    edges_progress = None

    if importlib.util.find_spec("tqdm") and show_progress_bar:
        from tqdm.auto import tqdm

        nodes_progress = tqdm(
            total=len(nodes), desc="uploading nodes, completed :: "
        )

    for node_id in nodes:
        node_attrs = __format_attrs(nx_graph.nodes[node_id]["attr"])
        node_type = node_attrs["type"]

        command = (
            "CREATE (n:" + node_type + " {id: $id})\nSET n += $properties"
        )

        session.run(command, id=node_id, properties=node_attrs)

        if nodes_progress is not None:
            nodes_progress.update()

    if nodes_progress is not None:
        nodes_progress.close()

    if importlib.util.find_spec("tqdm") and show_progress_bar:
        from tqdm.auto import tqdm

        edges_progress = tqdm(
            total=len(edges), desc="uploading edges, completed :: "
        )

    for edge_id in edges:
        edge_attr = __format_attrs(nx_graph.edges[edge_id]["attr"])
        edge_type = edge_attr["type"]

        command = (
            "MATCH (a {id: $id1}), (b {id: $id2})\nCREATE (a)-[r:"
            + edge_type
            + " $props]->(b)"
        )

        session.run(
            command,
            id1=edge_id[0],
            id2=edge_id[1],
            props=edge_attr,
            edge_type=edge_type,
        )

        if edges_progress is not None:
            edges_progress.update()

    if edges_progress is not None:
        edges_progress.close()


def neo4j_download(
    session, parameters: Optional[Dict[Any, Any]] = None
) -> nx.DiGraph:
    """
    Downloads a NetworkX DiGraph starting from a Neo4J database.

    Parameters
    --------------
    session
        Neo4J session
    parameters
        Optional parameters of the method.

    Returns
    --------------
    nx_graph
        NetworkX DiGraph
    """
    if parameters is None:
        parameters = {}

    from pm4py.util import dt_parsing

    date_parser = dt_parsing.parser.get()

    nodes = session.run("MATCH (n) RETURN n")
    nodes = [dict(node["n"]) for node in nodes]

    edges = session.run("MATCH (n)-[r]->(m) RETURN n, r, m")
    edges = [
        (edge["n"]["id"], edge["m"]["id"], dict(edge["r"])) for edge in edges
    ]

    nx_graph = DiGraph()

    for n in nodes:
        node_id = n["id"]
        node_props = copy(n)
        del node_props["id"]

        for k in ["ocel:timestamp", "time:timestamp"]:
            if k in node_props:
                node_props[k] = date_parser.apply(node_props[k])

        nx_graph.add_node(node_id, attr=node_props)

    for e in edges:
        nx_graph.add_edge(e[0], e[1], attr=e[2])

    return nx_graph


def nx_to_ocel(
    nx_graph: nx.DiGraph, parameters: Optional[Dict[Any, Any]] = None
):
    """
    Transforms a NetworkX DiGraph representing an OCEL to a proper OCEL.

    Parameters
    ----------------
    nx_graph
        NetworkX DiGraph
    parameters
        Optional parameters of the method

    Returns
    ----------------
    ocel
        Object-centric event log
    """
    if parameters is None:
        parameters = {}

    from pm4py.util import pandas_utils

    events = []
    objects = []
    relations = []
    o2o = []
    object_changes = []

    events_activity = {}
    events_timestamp = {}
    object_types = {}

    for node_id in nx_graph.nodes:
        node_attrs = nx_graph.nodes[node_id]["attr"]
        node_type = node_attrs["type"]

        if node_type == "EVENT":
            activity = node_attrs["ocel:activity"]
            timestamp = node_attrs["ocel:timestamp"]
            events_activity[node_id] = activity
            events_timestamp[node_id] = timestamp
            events.append(node_attrs)
        elif node_type == "OBJECT":
            object_type = node_attrs["ocel:type"]
            object_types[node_id] = object_type
            objects.append(node_attrs)
        elif node_type == "CHANGE":
            object_changes.append(node_attrs)

    for edge_id in nx_graph.edges:
        source = edge_id[0]
        target = edge_id[1]
        edge_attrs = nx_graph.edges[edge_id]["attr"]
        edge_type = edge_attrs["type"]
        qualifier = (
            edge_attrs["qualifier"] if "qualifier" in edge_attrs else ""
        )

        if edge_type == "E2O":
            activity = events_activity[source]
            timestamp = events_timestamp[source]
            object_type = object_types[target]
            relations.append(
                {
                    "ocel:eid": source,
                    "ocel:oid": target,
                    "ocel:activity": activity,
                    "ocel:timestamp": timestamp,
                    "ocel:type": object_type,
                    "ocel:qualifier": qualifier,
                }
            )
        elif edge_type == "O2O":
            o2o.append(
                {
                    "ocel:oid": source,
                    "ocel:oid_2": target,
                    "ocel:qualifier": qualifier,
                }
            )

    events = pandas_utils.instantiate_dataframe(events)
    objects = pandas_utils.instantiate_dataframe(objects)
    relations = pandas_utils.instantiate_dataframe(relations)
    o2o = pandas_utils.instantiate_dataframe(o2o) if o2o else None
    object_changes = (
        pandas_utils.instantiate_dataframe(object_changes)
        if object_changes
        else None
    )

    internal_index = "@@index"
    events = pandas_utils.insert_index(
        events, internal_index, reset_index=False, copy_dataframe=False
    )
    relations = pandas_utils.insert_index(
        relations, internal_index, reset_index=False, copy_dataframe=False
    )

    events = events.sort_values(["ocel:timestamp", internal_index])
    relations = relations.sort_values(["ocel:timestamp", internal_index])

    del events[internal_index]
    del relations[internal_index]

    del events["type"]
    del objects["type"]

    if object_changes is not None:
        del object_changes["type"]

    from pm4py.objects.ocel.obj import OCEL

    return OCEL(
        events, objects, relations, o2o=o2o, object_changes=object_changes
    )


def nx_to_event_log(
    nx_graph: nx.DiGraph, parameters: Optional[Dict[Any, Any]] = None
):
    """
    Transforms a NetworkX DiGraph representing a traditional event log to a proper event log.

    Parameters
    ----------------
    nx_graph
        NetworkX DiGraph
    parameters
        Optional parameters of the method

    Returns
    ----------------
    event_log
        Traditional event log.
    """
    if parameters is None:
        parameters = {}

    from pm4py.objects.log.obj import EventLog, Trace, Event
    from pm4py.objects.log.util import sorting

    log = EventLog()

    case_nodes = [
        (k, v["attr"])
        for k, v in nx_graph.nodes.items()
        if v["attr"]["type"] == "CASE"
    ]
    event_nodes = [
        (k, v["attr"])
        for k, v in nx_graph.nodes.items()
        if v["attr"]["type"] == "EVENT"
    ]

    cases = {}
    for i in range(len(case_nodes)):
        case_attrs = copy(case_nodes[i][1])
        del case_attrs["type"]
        trace = Trace(attributes=case_attrs)
        cases[case_nodes[i][0]] = trace
        log.append(trace)

    events = {}
    for i in range(len(event_nodes)):
        event_attrs = copy(event_nodes[i][1])
        del event_attrs["type"]
        events[event_nodes[i][0]] = Event(event_attrs)

    for edge_id, edge_attrs in nx_graph.edges.items():
        edge_attrs = edge_attrs["attr"]
        edge_type = edge_attrs["type"]

        if edge_type == "BELONGS_TO":
            cases[edge_id[1]].append(events[edge_id[0]])

    log = sorting.sort_timestamp(log, "time:timestamp")

    return log



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util import exec_utils
from enum import Enum
from pm4py.util import constants, xes_constants, pandas_utils
from pm4py.objects.conversion.log import converter
from pm4py.algo.discovery.correlation_mining.variants import classic
from collections import Counter
import numpy as np
from typing import Optional, Dict, Any, Union, Tuple
from pm4py.objects.log.obj import EventLog, EventStream
from pm4py.utils import is_polars_lazyframe
import pandas as pd
import importlib.util


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    SAMPLE_SIZE = "sample_size"


def apply(
    log: Union[EventLog, EventStream, pd.DataFrame],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Tuple[Dict[Tuple[str, str], int], Dict[Tuple[str, str], float]]:
    """
    Applies the correlation miner (splits the log in smaller chunks)

    Parameters
    ---------------
    log
        Log object
    parameters
        Parameters of the algorithm

    Returns
    ---------------
    dfg
        Frequency DFG
    performance_dfg
        Performance DFG
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    sample_size = exec_utils.get_param_value(
        Parameters.SAMPLE_SIZE, parameters, 100000
    )

    PS_matrixes = []
    duration_matrixes = []

    if pandas_utils.check_is_pandas_dataframe(log):
        if is_polars_lazyframe(log):
            if importlib.util.find_spec("polars") is None:
                raise RuntimeError(
                    "Polars LazyFrame provided but 'polars' package is not installed."
                )

            import polars as pl  # type: ignore[import-untyped]

            required_columns = list(
                {activity_key, timestamp_key, start_timestamp_key}
            )
            missing_columns = [
                col for col in required_columns if col not in log.columns
            ]
            if missing_columns:
                raise Exception(
                    "The provided Polars LazyFrame does not contain the following required columns: "
                    + ", ".join(sorted(missing_columns))
                )

            activities_counter = (
                log.group_by(activity_key)
                .agg(pl.len().alias("__pm4py_count__"))
                .collect()
            )
            activities_counter = {
                row[activity_key]: row["__pm4py_count__"]
                for row in activities_counter.iter_rows(named=True)
            }
            activities = sorted(list(activities_counter.keys()))

            log = (
                log.select(required_columns)
                .sort([timestamp_key, start_timestamp_key])
                .collect()
                .to_pandas()
            )
        else:
            # code for Pandas dataframes
            log = log[
                list(set([activity_key, timestamp_key, start_timestamp_key]))
            ]
            log = log.sort_values([timestamp_key, start_timestamp_key])
            activities_counter = log[activity_key].value_counts().to_dict()
            activities = sorted(list(activities_counter.keys()))
    else:
        log = converter.apply(
            log,
            variant=converter.Variants.TO_EVENT_STREAM,
            parameters={"deepcopy": False, "include_case_attributes": False},
        )
        activities_counter = Counter(x[activity_key] for x in log)
        activities = sorted(list(activities_counter.keys()))

    prev = 0
    while prev < len(log):
        sample = log[prev: min(len(log), prev + sample_size)]
        transf_stream, activities_grouped, activities = classic.preprocess_log(
            sample, activities=activities, parameters=parameters
        )
        PS_matrix, duration_matrix = classic.get_PS_dur_matrix(
            activities_grouped, activities, parameters=parameters
        )
        PS_matrixes.append(PS_matrix)
        duration_matrixes.append(duration_matrix)

        prev = prev + sample_size

    PS_matrix = np.zeros((len(activities), len(activities)))
    duration_matrix = np.zeros((len(activities), len(activities)))
    z = 0
    while z < len(PS_matrixes):
        PS_matrix = PS_matrix + PS_matrixes[z]
        duration_matrix = np.maximum(duration_matrix, duration_matrixes[z])
        z = z + 1
    PS_matrix = PS_matrix / float(len(PS_matrixes))

    return classic.resolve_lp_get_dfg(
        PS_matrix, duration_matrix, activities, activities_counter
    )



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util.constants import CASE_CONCEPT_NAME, PARAMETER_CONSTANT_CASEID_KEY, PARAMETER_CONSTANT_ACTIVITY_KEY
from enum import Enum
from pm4py.util import exec_utils, constants, xes_constants
from typing import Optional, Dict, Any, Union, List
import polars as pl


class Parameters(Enum):
    CASE_ID_KEY = PARAMETER_CONSTANT_CASEID_KEY
    ACTIVITY_KEY = PARAMETER_CONSTANT_ACTIVITY_KEY
    POSITIVE = "positive"


def apply(
    df: pl.LazyFrame,
    prefixes: List[List[str]],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Filter cases that start with a given prefix

    Parameters
    -----------
    df
        LazyFrame
    prefixes
        List of prefixes to match
    parameters
        Parameters of the algorithm

    Returns
    -----------
    df
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    positive = exec_utils.get_param_value(Parameters.POSITIVE, parameters, True)

    # Create variants
    variants_df = (
        df.sort([case_id_glue, "time:timestamp"])
        .group_by(case_id_glue, maintain_order=True)
        .agg(
            pl.col(activity_key)
            .str.concat(constants.DEFAULT_VARIANT_SEP)
            .alias("variant")
        )
    )

    # Convert prefixes to strings
    prefix_strings = [constants.DEFAULT_VARIANT_SEP.join(p) for p in prefixes]
    
    # Check if variant starts with any prefix
    prefix_expr = None
    for prefix in prefix_strings:
        if prefix_expr is None:
            prefix_expr = pl.col("variant").str.starts_with(prefix)
        else:
            prefix_expr = prefix_expr | pl.col("variant").str.starts_with(prefix)
    
    matching_cases = variants_df.filter(prefix_expr).select(case_id_glue)

    if positive:
        ret = df.join(matching_cases, on=case_id_glue, how="inner")
    else:
        ret = df.join(matching_cases, on=case_id_glue, how="anti")

    return ret



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Union, Optional, Dict, Any, List

import pandas as pd

from pm4py.algo.organizational_mining import util
from pm4py.objects.log.obj import EventLog
from pm4py.util import constants


class Parameters(Enum):
    RESOURCE_KEY = constants.PARAMETER_CONSTANT_RESOURCE_KEY
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    GROUP_KEY = constants.PARAMETER_CONSTANT_GROUP_KEY


class Outputs(Enum):
    GROUP_RELATIVE_FOCUS = "group_relative_focus"
    GROUP_RELATIVE_STAKE = "group_relative_stake"
    GROUP_COVERAGE = "group_coverage"
    GROUP_MEMBER_CONTRIBUTION = "group_member_contribution"


def apply_from_clustering_or_roles(
    log_obj: Union[pd.DataFrame, EventLog],
    ja_clustering_or_roles: Dict[str, List[str]],
    parameters: Optional[Dict[Any, str]] = None,
) -> Dict[str, Any]:
    """
    Provides the local diagnostics for the organizational model starting from a log object and the results
    of the similar activities clustering / the roles detection algorithm.

    The approach implemented is the one described in:
    Yang, Jing, et al. "OrgMining 2.0: A Novel Framework for Organizational Model Mining from Event Logs."
    arXiv preprint arXiv:2011.12445 (2020).

    Parameters
    --------------
    log_obj
        Log object
    ja_clustering_or_roles
        Result of the similar activities clustering / the roles detection algorithm
    parameters
        Parameters of the algorithm, including:
        - pm4py:param:resource_key => the resource attribute
        - pm4py:param:activity_key => the activity attribute
        - pm4py:param:group_key => the group

    Returns
    ---------------
    Dictionary containing four keys:
        - group_relative_focus => relative focus metric
        - group_relative_stake => relative stake metric
        - group_coverage => group coverage metric
        - group_member_contribution => group member contribution metric
    """
    if parameters is None:
        parameters = {}

    res_act, act_res = util.get_res_act_from_log(
        log_obj, parameters=parameters
    )
    resources = util.get_resources_from_log(log_obj, parameters=parameters)

    if type(ja_clustering_or_roles) is list:
        ja_clustering_or_roles = {
            str(i): ja_clustering_or_roles[i].originator_importance
            for i in range(len(ja_clustering_or_roles))
        }

    groups = {}
    for cluster in ja_clustering_or_roles:
        groups[cluster] = {}
        for res in ja_clustering_or_roles[cluster]:
            groups[cluster][res] = resources[res]

    return __apply(res_act, act_res, groups, parameters=parameters)


def apply_from_group_attribute(
    log_obj: Union[pd.DataFrame, EventLog],
    parameters: Optional[Dict[Any, str]] = None,
) -> Dict[str, Any]:
    """
    Provides the local diagnostics for the organizational model starting from a log object and considering
    the group specified by the attribute

    The approach implemented is the one described in:
    Yang, Jing, et al. "OrgMining 2.0: A Novel Framework for Organizational Model Mining from Event Logs."
    arXiv preprint arXiv:2011.12445 (2020).

    Parameters
    --------------
    log_obj
        Log object
    parameters
        Parameters of the algorithm, including:
        - pm4py:param:resource_key => the resource attribute
        - pm4py:param:activity_key => the activity attribute
        - pm4py:param:group_key => the group

    Returns
    ---------------
    Dictionary containing four keys:
        - group_relative_focus => relative focus metric
        - group_relative_stake => relative stake metric
        - group_coverage => group coverage metric
        - group_member_contribution => group member contribution metric
    """
    if parameters is None:
        parameters = {}

    res_act, act_res = util.get_res_act_from_log(
        log_obj, parameters=parameters
    )
    groups = util.get_groups_from_log(log_obj, parameters=parameters)
    return __apply(res_act, act_res, groups, parameters=parameters)


def __apply(
    res_act: Dict[str, Dict[str, int]],
    act_res: Dict[str, Dict[str, int]],
    groups: Dict[str, Dict[str, int]],
    parameters: Optional[Dict[Any, str]] = None,
) -> Dict[str, Any]:
    """
    Provides the local diagnostics for the organizational model

    The approach implemented is the one described in:
    Yang, Jing, et al. "OrgMining 2.0: A Novel Framework for Organizational Model Mining from Event Logs."
    arXiv preprint arXiv:2011.12445 (2020).

    Parameters
    ----------------
    res_act
        Dictionary resources-activities-occurrences
    act_res
        Dictionary activities-resources-occurrences
    groups
        Dictionary groups-resources-occurrences
    parameters
        Parameters of the algorithm

    Returns
    -----------------
    dict
        Dictionary containing four keys:
        - group_relative_focus => relative focus metric
        - group_relative_stake => relative stake metric
        - group_coverage => group coverage metric
        - group_member_contribution => group member contribution metric
    """
    if parameters is None:
        parameters = {}

    ret = {}
    ret[Outputs.GROUP_RELATIVE_FOCUS.value] = __group_relative_focus(
        res_act, act_res, groups, parameters=parameters
    )
    ret[Outputs.GROUP_RELATIVE_STAKE.value] = __group_relative_stake(
        res_act, act_res, groups, parameters=parameters
    )
    ret[Outputs.GROUP_COVERAGE.value] = __group_coverage(
        res_act, act_res, groups, parameters=parameters
    )
    ret[Outputs.GROUP_MEMBER_CONTRIBUTION.value] = __group_member_contribution(
        res_act, act_res, groups, parameters=parameters
    )

    return ret


def __group_relative_focus(
    res_act: Dict[str, Dict[str, int]],
    act_res: Dict[str, Dict[str, int]],
    groups: Dict[str, Dict[str, int]],
    parameters: Optional[Dict[Any, str]] = None,
) -> Dict[str, Dict[str, float]]:
    """
    Calculates the relative focus metric

    GROUP RELATIVE FOCUS (on a given type of work) specifies how much a resource group performed this type of work
    compared to the overall workload of the group. It can be used to measure how the workload of a resource group
    is distributed over different types of work, i.e., work diversification of the group.

    The approach implemented is the one described in:
    Yang, Jing, et al. "OrgMining 2.0: A Novel Framework for Organizational Model Mining from Event Logs."
    arXiv preprint arXiv:2011.12445 (2020).

    Parameters
    ----------------
    res_act
        Dictionary resources-activities-occurrences
    act_res
        Dictionary activities-resources-occurrences
    groups
        Dictionary groups-resources-occurrences
    parameters
        Parameters of the algorithm

    Returns
    -----------------
    metric
        Metric value
    """
    ret = {}
    for g in groups:
        res_in_g = groups[g]
        ret[g] = {}
        for a in act_res:
            this = 0
            total = 0
            for r in act_res[a]:
                total += act_res[a][r]
                if r in res_in_g:
                    this += act_res[a][r]
            ret[g][a] = float(this) / float(total)
    return ret


def __group_relative_stake(
    res_act: Dict[str, Dict[str, int]],
    act_res: Dict[str, Dict[str, int]],
    groups: Dict[str, Dict[str, int]],
    parameters: Optional[Dict[Any, str]] = None,
) -> Dict[str, Dict[str, float]]:
    """
    Calculates the relative stake metric

    GROUP RELATIVE STAKE (in a given type of work) specifies how much this type of work was performed by a certain
    resource group among all groups. It can be used to measure how the workload devoted to a certain type of work is
    distributed over resource groups in an organizational model, i.e., work participation by different groups.

    The approach implemented is the one described in:
    Yang, Jing, et al. "OrgMining 2.0: A Novel Framework for Organizational Model Mining from Event Logs."
    arXiv preprint arXiv:2011.12445 (2020).

    Parameters
    ----------------
    res_act
        Dictionary resources-activities-occurrences
    act_res
        Dictionary activities-resources-occurrences
    groups
        Dictionary groups-resources-occurrences
    parameters
        Parameters of the algorithm

    Returns
    -----------------
    metric
        Metric value
    """
    ret = {}
    for g in groups:
        total = 0
        res_in_g = groups[g]
        ret[g] = {}
        for a in act_res:
            this = 0
            for r in act_res[a]:
                if r in res_in_g:
                    total += act_res[a][r]
                    this += act_res[a][r]
            ret[g][a] = this
        for a in act_res:
            ret[g][a] = float(ret[g][a]) / float(total)
    return ret


def __group_coverage(
    res_act: Dict[str, Dict[str, int]],
    act_res: Dict[str, Dict[str, int]],
    groups: Dict[str, Dict[str, int]],
    parameters: Optional[Dict[Any, str]] = None,
) -> Dict[str, Dict[str, float]]:
    """
    Calculates the group coverage metric

    GROUP COVERAGE with respect to a given type of work specifies the proportion of members of a resource group that
    performed this type of work.

    The approach implemented is the one described in:
    Yang, Jing, et al. "OrgMining 2.0: A Novel Framework for Organizational Model Mining from Event Logs."
    arXiv preprint arXiv:2011.12445 (2020).

    Parameters
    ----------------
    res_act
        Dictionary resources-activities-occurrences
    act_res
        Dictionary activities-resources-occurrences
    groups
        Dictionary groups-resources-occurrences
    parameters
        Parameters of the algorithm

    Returns
    -----------------
    metric
        Metric value
    """
    ret = {}
    for g in groups:
        ret[g] = {}
        total = sum(groups[g].values())
        for r in groups[g]:
            ret[g][r] = float(groups[g][r]) / float(total)
    return ret


def __group_member_contribution(
    res_act: Dict[str, Dict[str, int]],
    act_res: Dict[str, Dict[str, int]],
    groups: Dict[str, Dict[str, int]],
    parameters: Optional[Dict[Any, str]] = None,
) -> Dict[str, Dict[str, Dict[str, int]]]:
    """
    Calculates the member contribution metric

    GROUP MEMBER CONTRIBUTION of a member of a resource group with respect to the given type of work specifies how
    much of this type of work by the group was performed by the member. It can be used to measure how the workload
    of the entire group devoted to a certain type of work is distributed over the group members.

    The approach implemented is the one described in:
    Yang, Jing, et al. "OrgMining 2.0: A Novel Framework for Organizational Model Mining from Event Logs."
    arXiv preprint arXiv:2011.12445 (2020).

    Parameters
    ----------------
    res_act
        Dictionary resources-activities-occurrences
    act_res
        Dictionary activities-resources-occurrences
    groups
        Dictionary groups-resources-occurrences
    parameters
        Parameters of the algorithm

    Returns
    -----------------
    metric
        Metric value
    """
    ret = {}
    for g in groups:
        ret[g] = {}
        for r in groups[g]:
            ret[g][r] = res_act[r]
    return ret



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from collections import defaultdict
from typing import Optional, Dict, Any, Tuple, Set
from pm4py.objects.ocel.obj import OCEL


def apply(ocel: OCEL,
          parameters: Optional[Dict[Any,
                                    Any]] = None) -> Tuple[Set[str],
                                                           Set[str],
                                                           Set[Tuple[str,
                                                                     str]],
                                                           Dict[Tuple[str,
                                                                      str],
                                                                int]]:
    """
    Discovers the ET-OT graph from an OCEL

    Published in: https://publications.rwth-aachen.de/record/1014107

    Parameters
    ---------------
    ocel
        Object-centric event log
    parameters
        Variant-specific parameters

    Returns
    ----------------
    activities
        Set of activities
    object_types
        Set of object types
    edges
        Set of edges
    edges_frequency
        Dictionary associating to each edge a frequency
    """
    if parameters is None:
        parameters = {}

    A = set()
    OT = set()
    R = set()
    w = defaultdict(int)

    # Iterate over the relations to build the ETOT Graph
    for index, row in ocel.relations.iterrows():
        a = row['ocel:activity']
        ot = row['ocel:type']
        A.add(a)
        OT.add(ot)
        R.add((a, ot))
        w[(a, ot)] += 1

    return A, OT, R, w



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import numpy as np

from pm4py.util import xes_constants, pandas_utils, constants
from pm4py.util.business_hours import soj_time_business_hours_diff
import pandas as pd


def get_dfg_graph(
        df,
        measure="frequency",
        activity_key="concept:name",
        case_id_glue="case:concept:name",
        start_timestamp_key=None,
        timestamp_key="time:timestamp",
        perf_aggregation_key="mean",
        sort_caseid_required=True,
        sort_timestamp_along_case_id=True,
        keep_once_per_case=False,
        window=1,
        business_hours=False,
        business_hours_slot=None,
        workcalendar=constants.DEFAULT_BUSINESS_HOURS_WORKCALENDAR,
        target_activity_key=None,
        reduce_columns=True,
        cost_attribute=None,
):
    """
    Get DFG graph from Pandas dataframe - optimized version

    Parameters
    -----------
    df
        Dataframe
    measure
        Measure to use (frequency/performance/both)
    activity_key
        Activity key to use in the grouping
    case_id_glue
        Case ID identifier
    start_timestamp_key
        Start timestamp key
    timestamp_key
        Timestamp key
    perf_aggregation_key
        Performance aggregation key (mean, median, min, max)
    sort_caseid_required
        Specify if a sort on the Case ID is required
    sort_timestamp_along_case_id
        Specifying if sorting by timestamp along the CaseID is required
    keep_once_per_case
        In the counts, keep only one occurrence of the path per case (the first)
    window
        Window of the DFG (default 1)

    Returns
    -----------
    dfg
        DFG in the chosen measure (may be only the frequency, only the performance, or both)
    """
    # added support to specify an activity key for the target event which is different
    # from the activity key of the source event.
    if target_activity_key is None:
        target_activity_key = activity_key

    # if not differently specified, set the start timestamp key to the timestamp key
    # to avoid retro-compatibility problems
    st_eq_ct = start_timestamp_key == timestamp_key
    if start_timestamp_key is None:
        start_timestamp_key = xes_constants.DEFAULT_START_TIMESTAMP_KEY
        if start_timestamp_key not in df.columns:
            df[start_timestamp_key] = df[timestamp_key]
        st_eq_ct = True

    # Determine which columns we need to keep before reducing
    needed_columns = {case_id_glue, activity_key, target_activity_key}
    if measure != "frequency" or sort_timestamp_along_case_id:
        needed_columns.update({start_timestamp_key, timestamp_key})
    if measure == "cost":
        needed_columns.add(cost_attribute)

    # Reduce dataframe to necessary columns (more efficient than previous implementation)
    if reduce_columns:
        df = df[list(needed_columns)]

    if measure == "cost":
        df[cost_attribute] = df[cost_attribute].fillna(0)  # Using 0 directly is more efficient

    # Create a view with the needed sorting for more efficient memory usage
    if sort_caseid_required:
        if sort_timestamp_along_case_id:
            df = df.sort_values([case_id_glue, start_timestamp_key, timestamp_key])
        else:
            df = df.sort_values(case_id_glue)

    # Pre-create the suffixed column names for more efficient joining
    suffix = "_2"
    # Get a list of column names before shift to avoid recreating them
    orig_cols = df.columns.tolist()
    shifted_cols = [str(col) + suffix for col in orig_cols]

    # Shift the dataframe by window
    df_shifted = df.shift(-window)
    df_shifted.columns = shifted_cols

    # Concatenate the dataframes efficiently
    df_successive_rows = pandas_utils.DATAFRAME.concat([df, df_shifted], axis=1)

    # Filter for matching case IDs (more efficient with direct access)
    case_id_col = case_id_glue
    case_id_col_shifted = case_id_glue + suffix
    df_successive_rows = df_successive_rows[df_successive_rows[case_id_col] == df_successive_rows[case_id_col_shifted]]

    # Handle keep_once_per_case more efficiently
    if keep_once_per_case:
        # Use drop_duplicates which is more efficient than groupby + first for this operation
        df_successive_rows = df_successive_rows.drop_duplicates(
            subset=[case_id_glue, activity_key, target_activity_key + suffix]
        )

    # Calculate performance metrics if needed
    if measure in ["performance", "both"]:
        if not st_eq_ct:
            # Calculate max more efficiently using numpy
            df_successive_rows[start_timestamp_key + suffix] = np.maximum(
                df_successive_rows[start_timestamp_key + suffix],
                df_successive_rows[timestamp_key]
            )

        # Calculate the time difference
        if business_hours:
            if business_hours_slot is None:
                business_hours_slot = constants.DEFAULT_BUSINESS_HOUR_SLOTS

            # Business hours calculation requires Python datetime objects
            # Use a more efficient approach than apply, but ensure we're working with correct datetime types

            # Convert timestamps - we need to use to_pydatetime() to get Python datetime objects
            # that are compatible with the business hours function
            if hasattr(df_successive_rows[timestamp_key], 'dt'):
                # If timestamps are already pandas datetime, convert to Python datetime
                ts_values = np.array(df_successive_rows[timestamp_key].dt.to_pydatetime())
                start_ts_values = np.array(df_successive_rows[start_timestamp_key + suffix].dt.to_pydatetime())
            else:
                # If timestamps are pandas Timestamp objects or strings, convert properly
                ts_values = np.array(pandas_utils.DATAFRAME.to_datetime(df_successive_rows[timestamp_key]).dt.to_pydatetime())
                start_ts_values = np.array(pandas_utils.DATAFRAME.to_datetime(df_successive_rows[start_timestamp_key + suffix]).dt.to_pydatetime())

            # Use list comprehension which is faster than apply but handles datetime objects correctly
            flow_times = [
                soj_time_business_hours_diff(ts, start_ts, business_hours_slot, workcalendar)
                for ts, start_ts in zip(ts_values, start_ts_values)
            ]

            df_successive_rows[constants.DEFAULT_FLOW_TIME] = flow_times
        else:
            # Vectorized timestamp difference calculation (already efficient)
            difference = df_successive_rows[start_timestamp_key + suffix] - df_successive_rows[timestamp_key]
            df_successive_rows[constants.DEFAULT_FLOW_TIME] = pandas_utils.get_total_seconds(difference)

    # Set up grouping based on measure
    group_cols = [activity_key, target_activity_key + suffix]

    if measure == "performance" or measure == "both":
        agg_col = constants.DEFAULT_FLOW_TIME
    elif measure == "cost":
        agg_col = cost_attribute + suffix
    else:
        # For frequency, find the first non-grouping column to use (more efficient)
        remaining_cols = set(df_successive_rows.columns) - set(group_cols)
        agg_col = next(iter(remaining_cols)) if remaining_cols else case_id_col

    # Create dictionary directly rather than creating a Series object first (more efficient)
    if measure == "frequency" or measure == "both":
        # Use value_counts which is more efficient than groupby+size for frequency
        if type(df) is pd.DataFrame and len(group_cols) == 2:  # Most common case
            temp_df = df_successive_rows[group_cols].copy()
            temp_df['dummy'] = 1  # Add a column to count
            pivot = temp_df.pivot_table(
                index=group_cols[0],
                columns=group_cols[1],
                values='dummy',
                aggfunc='count',
                fill_value=0
            )
            dfg_frequency = {(idx, col): val for idx, row in pivot.iterrows() for col, val in row.items() if val > 0}
        else:
            # Fallback to groupby
            dfg_frequency = df_successive_rows.groupby(group_cols).size().to_dict()

    # Performance metrics calculation
    dfg_performance = {}
    if measure in ["performance", "cost", "both"]:
        grouped = df_successive_rows.groupby(group_cols)[agg_col]

        if perf_aggregation_key == "all":
            # Calculate all metrics at once (more efficient than separate calls)
            metrics = grouped.agg(['mean', 'median', 'max', 'min', 'sum', 'std'])

            if type(metrics) is pd.DataFrame:
                metrics = metrics.iterrows()
            else:
                metrics = metrics.to_pandas().iterrows()

            # Convert to the expected dictionary structure
            dfg_performance = {
                group: {
                    'mean': row['mean'],
                    'median': row['median'],
                    'max': row['max'],
                    'min': row['min'],
                    'sum': row['sum'],
                    'stdev': row['std']
                }
                for group, row in metrics
            }
        elif perf_aggregation_key == "raw_values":
            dfg_performance = grouped.agg(list).to_dict()
        else:
            dfg_performance = grouped.agg(perf_aggregation_key).to_dict()

    # Return appropriate results based on measure
    if measure == "frequency":
        return dfg_frequency
    elif measure == "performance" or measure == "cost":
        return dfg_performance
    else:  # measure == "both"
        return [dfg_frequency, dfg_performance]


def get_partial_order_dataframe(
    df,
    start_timestamp_key=None,
    timestamp_key="time:timestamp",
    case_id_glue="case:concept:name",
    activity_key="concept:name",
    sort_caseid_required=True,
    sort_timestamp_along_case_id=True,
    reduce_dataframe=True,
    keep_first_following=True,
    business_hours=False,
    business_hours_slot=None,
    workcalendar=constants.DEFAULT_BUSINESS_HOURS_WORKCALENDAR,
    event_index=constants.DEFAULT_INDEX_KEY,
):
    """
    Gets the partial order between events (of the same case) in a Pandas dataframe

    Parameters
    --------------
    df
        Dataframe
    start_timestamp_key
        Start timestamp key (if not provided, defaulted to the timestamp_key)
    timestamp_key
        Complete timestamp
    case_id_glue
        Column of the dataframe to use as case ID
    activity_key
        Activity key
    sort_caseid_required
        Tells if a sort by case ID is required (default: True)
    sort_timestamp_along_case_id
        Tells if a sort by timestamp is required along the case ID (default: True)
    reduce_dataframe
        To fasten operation, keep only essential columns in the dataframe
    keep_first_following
        Keep only the first event following the given event
    Returns
    ---------------
    part_ord_dataframe
        Partial order dataframe (with @@flow_time between events)
    """
    # if not differently specified, set the start timestamp key to the timestamp key
    # to avoid retro-compatibility problems
    if start_timestamp_key is None:
        start_timestamp_key = xes_constants.DEFAULT_START_TIMESTAMP_KEY

    if start_timestamp_key not in df:
        df[start_timestamp_key] = df[timestamp_key]

    # to increase the speed of the approaches reduce dataframe to case, activity (and possibly complete timestamp)
    # columns
    if reduce_dataframe:
        needed_columns = {
            case_id_glue,
            activity_key,
            start_timestamp_key,
            timestamp_key,
        }
        if event_index in df.columns:
            needed_columns.add(event_index)
        needed_columns = list(needed_columns)
        df = df[needed_columns]

    # to get rows belonging to same case ID together, we need to sort on case
    # ID
    if sort_caseid_required:
        if sort_timestamp_along_case_id:
            df = df.sort_values(
                [case_id_glue, start_timestamp_key, timestamp_key]
            )
        else:
            df = df.sort_values(case_id_glue)
        df = df.reset_index(drop=True)

    if event_index not in df.columns:
        df = pandas_utils.insert_index(
            df, event_index, copy_dataframe=False, reset_index=False
        )

    df = df.set_index(case_id_glue)

    df = df.join(df, rsuffix="_2")
    df = df[df[event_index] < df[event_index + "_2"]]
    df = df[df[timestamp_key] <= df[start_timestamp_key + "_2"]]

    df = df.reset_index()

    if business_hours:
        if business_hours_slot is None:
            business_hours_slot = constants.DEFAULT_BUSINESS_HOUR_SLOTS
        df[constants.DEFAULT_FLOW_TIME] = df.apply(
            lambda x: soj_time_business_hours_diff(
                x[timestamp_key],
                x[start_timestamp_key + "_2"],
                business_hours_slot,
                workcalendar,
            ),
            axis=1,
        )
    else:
        df[constants.DEFAULT_FLOW_TIME] = pandas_utils.get_total_seconds(
            df[start_timestamp_key + "_2"] - df[timestamp_key]
        )

    if keep_first_following:
        df = df.groupby(constants.DEFAULT_INDEX_KEY).first().reset_index()

    return df


def get_concurrent_events_dataframe(
    df,
    start_timestamp_key=None,
    timestamp_key="time:timestamp",
    case_id_glue="case:concept:name",
    activity_key="concept:name",
    sort_caseid_required=True,
    sort_timestamp_along_case_id=True,
    reduce_dataframe=True,
    max_start_column="@@max_start_column",
    min_complete_column="@@min_complete_column",
    diff_maxs_minc="@@diff_maxs_minc",
    strict=False,
):
    """
    Gets the concurrent events (of the same case) in a Pandas dataframe

    Parameters
    --------------
    df
        Dataframe
    start_timestamp_key
        Start timestamp key (if not provided, defaulted to the timestamp_key)
    timestamp_key
        Complete timestamp
    case_id_glue
        Column of the dataframe to use as case ID
    activity_key
        Activity key
    sort_caseid_required
        Tells if a sort by case ID is required (default: True)
    sort_timestamp_along_case_id
        Tells if a sort by timestamp is required along the case ID (default: True)
    reduce_dataframe
        To fasten operation, keep only essential columns in the dataframe
    strict
        Gets only entries that are strictly concurrent (i.e. the length of the intersection as real interval is > 0)

    Returns
    ---------------
    conc_ev_dataframe
        Concurrent events dataframe (with @@diff_maxs_minc as the size of the intersection of the intervals)
    """
    # if not differently specified, set the start timestamp key to the timestamp key
    # to avoid retro-compatibility problems
    if start_timestamp_key is None:
        start_timestamp_key = xes_constants.DEFAULT_START_TIMESTAMP_KEY
        df[start_timestamp_key] = df[timestamp_key]

    # to get rows belonging to same case ID together, we need to sort on case
    # ID
    if sort_caseid_required:
        if sort_timestamp_along_case_id:
            df = df.sort_values(
                [case_id_glue, start_timestamp_key, timestamp_key]
            )
        else:
            df = df.sort_values(case_id_glue)

    # to increase the speed of the approaches reduce dataframe to case, activity (and possibly complete timestamp)
    # columns
    if reduce_dataframe:
        df = df[
            [case_id_glue, activity_key, start_timestamp_key, timestamp_key]
        ]

    df = pandas_utils.insert_index(df)
    df = df.set_index(case_id_glue)
    df_copy = df.copy()

    df = df.join(df_copy, rsuffix="_2").dropna()
    df = df[
        df[constants.DEFAULT_INDEX_KEY]
        < df[constants.DEFAULT_INDEX_KEY + "_2"]
    ]
    df[max_start_column] = df[
        [start_timestamp_key, start_timestamp_key + "_2"]
    ].max(axis=1)
    df[min_complete_column] = df[[timestamp_key, timestamp_key + "_2"]].min(
        axis=1
    )
    df[max_start_column] = df[max_start_column].apply(lambda x: x.timestamp())
    df[min_complete_column] = df[min_complete_column].apply(
        lambda x: x.timestamp()
    )
    df[diff_maxs_minc] = df[min_complete_column] - df[max_start_column]
    if strict:
        df = df[df[diff_maxs_minc] > 0]
    else:
        df = df[df[diff_maxs_minc] >= 0]

    return df
